import datetime
import json
import traceback
import numpy as np

class AcqEngMetadata:

    CHANNEL_GROUP = "ChannelGroup"
    CORE_AUTOFOCUS_DEVICE = "Core-Autofocus"
    CORE_CAMERA = "Core-Camera"
    CORE_GALVO = "Core-Galvo"
    CORE_IMAGE_PROCESSOR = "Core-ImageProcessor"
    CORE_SLM = "Core-SLM"
    CORE_SHUTTER = "Core-Shutter"
    WIDTH = "Width"
    HEIGHT = "Height"
    PIX_SIZE = "PixelSize_um"
    POS_NAME = "PositionName"
    X_UM_INTENDED = "XPosition_um_Intended"
    Y_UM_INTENDED = "YPosition_um_Intended"
    Z_UM_INTENDED = "ZPosition_um_Intended"
    GENERIC_UM_INTENDED_SUFFIX = "Position_um_Intended"
    X_UM = "XPosition_um"
    Y_UM = "YPosition_um"
    Z_UM = "ZPosition_um"
    EXPOSURE = "Exposure"
    CHANNEL_NAME = "Channel"
    ZC_ORDER = "SlicesFirst"  # this is called ZCT in the functions
    TIME = "Time"
    DATE_TIME = "DateAndTime"
    SAVING_PREFIX = "Prefix"
    INITIAL_POS_LIST = "InitialPositionList"
    TIMELAPSE_INTERVAL = "Interval_ms"
    PIX_TYPE = "PixelType"
    BIT_DEPTH = "BitDepth"
    ELAPSED_TIME_MS = "ElapsedTime-ms"
    Z_STEP_UM = "z-step_um"
    EXPLORE_ACQUISITION = "ExploreAcquisition"
    AXES_GRID_COL = "column"
    AXES_GRID_ROW = "row"
    OVERLAP_X = "GridPixelOverlapX"
    OVERLAP_Y = "GridPixelOverlapY"
    AFFINE_TRANSFORM = "AffineTransform"
    PIX_TYPE_GRAY8 = "GRAY8"
    PIX_TYPE_GRAY16 = "GRAY16"
    CORE_XYSTAGE = "Core-XYStage"
    CORE_FOCUS = "Core-Focus"
    AXES = "Axes"
    CHANNEL_AXIS = "channel"
    TIME_AXIS = "time"
    Z_AXIS = "z"
    POSITION_AXIS = "position"
    TAGS = "tags"
    ACQUISITION_EVENT = "Event"

    @staticmethod
    def add_image_metadata(core, tags, event, elapsed_ms, exposure):
        try:
            AcqEngMetadata.set_pixel_size_um(tags, core.get_pixel_size_um())

            # Date and time
            AcqEngMetadata.set_elapsed_time_ms(tags, elapsed_ms)
            AcqEngMetadata.set_image_time(tags, datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S -'))

            # Info about all hardware that the core specifically knows about
            AcqEngMetadata.create_axes(tags)

            # Axes positions
            for s in event.get_defined_axes():
                AcqEngMetadata.set_axis_position(tags, s, event.get_axis_position(s))

            # XY Stage Positions
            if event.get_x_position() is not None and event.get_y_position() is not None:
                AcqEngMetadata.set_stage_x_intended(tags, event.get_x_position())
                AcqEngMetadata.set_stage_y_intended(tags, event.get_y_position())
            if event.get_position_name() is not None:
                AcqEngMetadata.set_position_name(tags, event.get_position_name())

            if event.get_z_position() is not None:
                AcqEngMetadata.set_stage_z_intended(tags, event.get_z_position())
            elif event.get_stage_single_axis_stage_position(core.get_focus_device()) is not None:
                AcqEngMetadata.set_stage_z_intended(tags,
                                                    event.get_stage_single_axis_stage_position(core.get_focus_device()))

            for name in event.get_stage_device_names():
                if name != core.get_focus_device():
                    AcqEngMetadata.set_stage_position_intended(tags, name,
                                                               event.get_stage_single_axis_stage_position(name))

            if event.get_sequence() is not None:
                AcqEngMetadata.add_acquisition_event(tags, event)

            AcqEngMetadata.set_exposure(tags, exposure)

        except Exception as e:
            traceback.print_exc()
            raise RuntimeError("Problem adding image metadata")

    @staticmethod
    def add_acquisition_event(tags, event):
        tags[AcqEngMetadata.ACQUISITION_EVENT] = event.toJSON()


    @staticmethod
    def make_summary_metadata(core, acq):
        summary = json.loads("{}")

        AcqEngMetadata.set_acq_date(summary, AcqEngMetadata.get_current_date_and_time())

        # General information the core-camera
        byte_depth = int(core.get_bytes_per_pixel())
        if byte_depth == 0:
            raise RuntimeError("Camera byte depth cannot be zero")
        AcqEngMetadata.set_pixel_type_from_byte_depth(summary, byte_depth)
        AcqEngMetadata.set_pixel_size_um(summary, core.get_pixel_size_um())

        # Info about core devices
        try:
            AcqEngMetadata.set_core_xy(summary, core.get_xy_stage_device())
            AcqEngMetadata.set_core_focus(summary, core.get_focus_device())
            AcqEngMetadata.set_core_autofocus(summary, core.get_auto_focus_device())
            AcqEngMetadata.set_core_camera(summary, core.get_camera_device())
            AcqEngMetadata.set_core_galvo(summary, core.get_galvo_device())
            AcqEngMetadata.set_core_image_processor(summary, core.get_image_processor_device())
            AcqEngMetadata.set_core_slm(summary, core.get_slm_device())
            AcqEngMetadata.set_core_shutter(summary, core.get_shutter_device())
        except Exception as e:
            raise RuntimeError("couldn't get info from core about devices")

        # TODO restore
        # # Affine transform
        # if AffineTransformUtils.isAffineTransformDefined():
        #     at = AffineTransformUtils.getAffineTransform(0, 0)
        #     AcqEngMetadata.setAffineTransformString(summary, AffineTransformUtils.transformToString(at))
        # else:
        #     AcqEngMetadata.setAffineTransformString(summary, "Undefined")

        return summary
    

    @staticmethod
    def get_current_date_and_time():
        return datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S")

    @staticmethod
    def get_indices(image_label):
        s = image_label.split("_")
        return [int(i) for i in s]

    @staticmethod
    def copy(map):
        return json.loads(json.dumps(map))

    @staticmethod
    def set_core_xy(map, xy_name):
        map[AcqEngMetadata.CORE_XYSTAGE] = xy_name

    @staticmethod
    def has_core_xy(map):
        return AcqEngMetadata.CORE_XYSTAGE in map

    @staticmethod
    def get_core_xy(map):
        if AcqEngMetadata.CORE_XYSTAGE in map:
            return map[AcqEngMetadata.CORE_XYSTAGE]
        else:
            raise ValueError("Missing core xy stage tag")

    @staticmethod
    def set_core_focus(map, z_name):
        map[AcqEngMetadata.CORE_FOCUS] = z_name

    @staticmethod
    def has_core_focus(map):
        return AcqEngMetadata.CORE_FOCUS in map

    @staticmethod
    def get_core_focus(map):
        if AcqEngMetadata.CORE_FOCUS in map:
            return map[AcqEngMetadata.CORE_FOCUS]
        else:
            raise ValueError("Missing core focus tag")

    @staticmethod
    def set_acq_date(map, date_time):
        map[AcqEngMetadata.DATE_TIME] = date_time

    @staticmethod
    def is_explore_acq(summary_metadata):
        if AcqEngMetadata.EXPLORE_ACQUISITION in summary_metadata:
            return summary_metadata[AcqEngMetadata.EXPLORE_ACQUISITION]
        else:
            raise ValueError("Missing explore tag")

    @staticmethod
    def set_explore_acq(summary_metadata, b):
        summary_metadata[AcqEngMetadata.EXPLORE_ACQUISITION] = b

    @staticmethod
    def has_acq_date(map):
        return AcqEngMetadata.DATE_TIME in map

    @staticmethod
    def get_acq_date(map):
        if AcqEngMetadata.DATE_TIME in map:
            return map[AcqEngMetadata.DATE_TIME]
        else:
            raise ValueError("Missing Acq dat time tag")

    @staticmethod
    def set_bit_depth(map, bit_depth):
        map[AcqEngMetadata.BIT_DEPTH] = bit_depth

    @staticmethod
    def has_bit_depth(map):
        return AcqEngMetadata.BIT_DEPTH in map

    @staticmethod
    def get_bit_depth(map):
        try:
            return map[AcqEngMetadata.BIT_DEPTH]
        except KeyError:
            raise ValueError("Missing bit depth tag")

    @staticmethod
    def set_width(map, width):
        map[AcqEngMetadata.WIDTH] = width

    @staticmethod
    def has_width(map):
        return AcqEngMetadata.WIDTH in map

    @staticmethod
    def get_width(map):
        try:
            return map[AcqEngMetadata.WIDTH]
        except KeyError:
            raise ValueError("Image width tag missing")

    @staticmethod
    def set_height(map, height):
        map[AcqEngMetadata.HEIGHT] = height

    @staticmethod
    def has_height(map):
        return AcqEngMetadata.HEIGHT in map

    @staticmethod
    def get_height(map):
        try:
            return map[AcqEngMetadata.HEIGHT]
        except KeyError:
            raise ValueError("Height missing from image tags")

    @staticmethod
    def set_position_name(map, position_name):
        map[AcqEngMetadata.POS_NAME] = position_name

    @staticmethod
    def has_position_name(map):
        return AcqEngMetadata.POS_NAME in map

    @staticmethod
    def get_position_name(map):
        try:
            return map[AcqEngMetadata.POS_NAME]
        except KeyError:
            raise ValueError("Missing position name tag")

    @staticmethod
    def set_pixel_type_from_string(map, pixel_type):
        map[AcqEngMetadata.PIX_TYPE] = pixel_type

    @staticmethod
    def set_pixel_type_from_byte_depth(map, depth):
        try:
            if depth == 1:
                map[AcqEngMetadata.PIX_TYPE] = AcqEngMetadata.PIX_TYPE_GRAY8
            elif depth == 2:
                map[AcqEngMetadata.PIX_TYPE] = AcqEngMetadata.PIX_TYPE_GRAY16
            elif depth == 4:
                map[AcqEngMetadata.PIX_TYPE] = AcqEngMetadata.PIX_TYPE_RGB32
        except KeyError:
            raise ValueError("Couldn't set pixel type")

    @staticmethod
    def has_pixel_type(map):
        return AcqEngMetadata.PIX_TYPE in map

    @staticmethod
    def get_pixel_type(map):
        try:
            return map[AcqEngMetadata.PIX_TYPE]
        except KeyError:
            raise ValueError("Missing pixel type tag")

    @staticmethod
    def get_bytes_per_pixel(map):
        if AcqEngMetadata.is_gray8(map):
            return 1
        elif AcqEngMetadata.is_gray16(map):
            return 2
        elif AcqEngMetadata.is_rgb32(map):
            return 4
        else:
            return 0

    @staticmethod
    def get_number_of_components(map):
        pixel_type = AcqEngMetadata.get_pixel_type(map)
        if pixel_type == AcqEngMetadata.PIX_TYPE_GRAY8 or pixel_type == AcqEngMetadata.PIX_TYPE_GRAY16:
            return 1
        elif pixel_type == AcqEngMetadata.PIX_TYPE_RGB32:
            return 3
        else:
            raise ValueError("Invalid pixel type")

    @staticmethod
    def is_gray8(map):
        return AcqEngMetadata.get_pixel_type(map) == AcqEngMetadata.PIX_TYPE_GRAY8

    @staticmethod
    def is_gray16(map):
        return AcqEngMetadata.get_pixel_type(map) == AcqEngMetadata.PIX_TYPE_GRAY16

    @staticmethod
    def is_rgb32(map):
        return AcqEngMetadata.get_pixel_type(map) == AcqEngMetadata.PIX_TYPE_RGB32

    @staticmethod
    def is_gray(map):
        return AcqEngMetadata.is_gray8(map) or AcqEngMetadata.is_gray16(map)

    @staticmethod
    def is_rgb(map):
        return AcqEngMetadata.is_rgb32(map)

    @staticmethod
    def get_keys(md):
        n = len(md)
        key_array = [None] * n
        keys = md.keys()
        for i in range(n):
            key_array[i] = keys.next()
        return key_array

    @staticmethod
    def get_json_array_member(obj, key):
        try:
            return obj[key]
        except KeyError:
            raise ValueError("Missing JSONArray member")

    @staticmethod
    def set_image_time(map, time):
        try:
            map[AcqEngMetadata.TIME] = time
        except KeyError:
            raise ValueError("Couldn't set image time")

    @staticmethod
    def has_image_time(map):
        return AcqEngMetadata.TIME in map

    @staticmethod
    def get_image_time(map):
        try:
            return map[AcqEngMetadata.TIME]
        except KeyError:
            raise ValueError("Missing image time tag")

    @staticmethod
    def get_depth(tags):
        pixel_type = AcqEngMetadata.get_pixel_type(tags)
        if AcqEngMetadata.PIX_TYPE_GRAY8 in pixel_type:
            return 1
        elif AcqEngMetadata.PIX_TYPE_GRAY16 in pixel_type:
            return 2
        else:
            return 0

    @staticmethod
    def set_exposure(map, exp):
        try:
            map[AcqEngMetadata.EXPOSURE] = exp
        except KeyError:
            raise ValueError("Could not set exposure")

    @staticmethod
    def has_exposure(map):
        return AcqEngMetadata.EXPOSURE in map

    @staticmethod
    def get_exposure(map):
        try:
            return map[AcqEngMetadata.EXPOSURE]
        except KeyError:
            raise ValueError("Exposure tag missing")

    @staticmethod
    def set_pixel_size_um(map, val):
        try:
            map[AcqEngMetadata.PIX_SIZE] = val
        except KeyError:
            raise ValueError("Missing pixel size tag")

    @staticmethod
    def has_pixel_size_um(map):
        return AcqEngMetadata.PIX_SIZE in map

    @staticmethod
    def get_pixel_size_um(map):
        try:
            return map[AcqEngMetadata.PIX_SIZE]
        except KeyError:
            raise ValueError("Pixel size missing in metadata")

    @staticmethod
    def set_z_step_um(map, val):
        try:
            map[AcqEngMetadata.Z_STEP_UM] = val
        except KeyError:
            raise ValueError("Couldn't set z step tag")

    @staticmethod
    def has_z_step_um(map):
        return AcqEngMetadata.Z_STEP_UM in map

    @staticmethod
    def get_z_step_um(map):
        try:
            return map[AcqEngMetadata.Z_STEP_UM]
        except KeyError:
            raise ValueError("Z step metadata field missing")

    @staticmethod
    def set_z_position_um(map, val):
        try:
            map[AcqEngMetadata.Z_UM] = val
        except KeyError:
            raise ValueError("Couldn't set z position")

    @staticmethod
    def has_z_position_um(map):
        return AcqEngMetadata.Z_UM in map

    @staticmethod
    def get_z_position_um(map):
        try:
            return map[AcqEngMetadata.Z_UM]
        except KeyError:
            raise ValueError("Missing Z position tag")

    @staticmethod
    def set_elapsed_time_ms(map, val):
        try:
            map[AcqEngMetadata.ELAPSED_TIME_MS] = val
        except KeyError:
            raise ValueError("Couldn't set elapsed time")

    @staticmethod
    def has_elapsed_time_ms(map):
        return AcqEngMetadata.ELAPSED_TIME_MS in map

    @staticmethod
    def get_elapsed_time_ms(map):
        try:
            return map[AcqEngMetadata.ELAPSED_TIME_MS]
        except KeyError:
            raise RuntimeError("missing elapsed time tag")

    @staticmethod
    def set_interval_ms(map, val):
        map[AcqEngMetadata.TIMELAPSE_INTERVAL] = val

    @staticmethod
    def has_interval_ms(map):
        return AcqEngMetadata.TIMELAPSE_INTERVAL in map

    @staticmethod
    def get_interval_ms(map):
        try:
            return map[AcqEngMetadata.TIMELAPSE_INTERVAL]
        except KeyError:
            raise RuntimeError("Time interval missing from summary metadata")

    @staticmethod
    def set_zct_order(map, val):
        map[AcqEngMetadata.ZC_ORDER] = val

    @staticmethod
    def has_zct_order(map):
        return AcqEngMetadata.ZC_ORDER in map

    @staticmethod
    def get_zct_order(map):
        try:
            return map[AcqEngMetadata.ZC_ORDER]
        except KeyError:
            raise RuntimeError("Missing ZCT Tag")

    @staticmethod
    def set_affine_transform_string(summary_md, affine):
        summary_md[AcqEngMetadata.AFFINE_TRANSFORM] = affine

    @staticmethod
    def has_affine_transform_string(map):
        return AcqEngMetadata.AFFINE_TRANSFORM in map

    @staticmethod
    def get_affine_transform_string(summary_md):
        try:
            return summary_md[AcqEngMetadata.AFFINE_TRANSFORM]
        except KeyError:
            raise RuntimeError("Affine transform missing from summary metadata")

    @staticmethod
    def get_affine_transform(summary_md):
        try:
            return AcqEngMetadata.string_to_transform(summary_md[AcqEngMetadata.AFFINE_TRANSFORM])
        except KeyError:
            raise RuntimeError("Affine transform missing from summary metadata")

    @staticmethod
    def string_to_transform(s):
        if s == "Undefined":
            return None
        mat = [0] * 4
        vals = s.split("_")
        for i in range(4):
            mat[i] = float(vals[i])
        return AcqEngMetadata.AffineTransform(mat)

    @staticmethod
    def set_pixel_overlap_x(smd, overlap):
        smd[AcqEngMetadata.OVERLAP_X] = overlap

    @staticmethod
    def has_pixel_overlap_x(map):
        return AcqEngMetadata.OVERLAP_X in map

    @staticmethod
    def get_pixel_overlap_x(summary_md):
        try:
            return summary_md[AcqEngMetadata.OVERLAP_X]
        except KeyError:
            raise RuntimeError("Could not find pixel overlap in image tags")

    @staticmethod
    def set_pixel_overlap_y(smd, overlap):
        smd[AcqEngMetadata.OVERLAP_Y] = overlap

    @staticmethod
    def has_pixel_overlap_y(map):
        return AcqEngMetadata.OVERLAP_Y in map

    @staticmethod
    def get_pixel_overlap_y(summary_md):
        try:
            return summary_md[AcqEngMetadata.OVERLAP_Y]
        except KeyError:
            raise RuntimeError("Could not find pixel overlap in image tags")

    @staticmethod
    def set_stage_x_intended(smd, x):
        smd[AcqEngMetadata.X_UM_INTENDED] = x

    @staticmethod
    def has_stage_x_intended(map):
        return AcqEngMetadata.X_UM_INTENDED in map

    @staticmethod
    def get_stage_x_intended(smd):
        try:
            return smd[AcqEngMetadata.X_UM_INTENDED]
        except KeyError:
            raise RuntimeError("Could not get stage x")

    @staticmethod
    def set_stage_y_intended(smd, y):
        smd[AcqEngMetadata.Y_UM_INTENDED] = y

    @staticmethod
    def has_stage_y_intended(map):
        return AcqEngMetadata.Y_UM_INTENDED in map

    @staticmethod
    def get_stage_y_intended(smd):
        try:
            return smd[AcqEngMetadata.Y_UM_INTENDED]
        except KeyError:
            raise RuntimeError("Could not get stage y")

    @staticmethod
    def set_stage_z_intended(smd, y):
        smd[AcqEngMetadata.Z_UM_INTENDED] = y

    @staticmethod
    def has_stage_z_intended(map):
        return AcqEngMetadata.Z_UM_INTENDED in map

    @staticmethod
    def get_stage_z_intended(smd):
        try:
            return smd[AcqEngMetadata.Z_UM_INTENDED]
        except KeyError:
            raise RuntimeError("Could not get stage Z")

    @staticmethod
    def set_stage_position_intended(tags, name, stage_single_axis_stage_position):
        tags[name + AcqEngMetadata.GENERIC_UM_INTENDED_SUFFIX] = stage_single_axis_stage_position

    @staticmethod
    def set_stage_x(smd, x):
        smd[AcqEngMetadata.X_UM] = x

    @staticmethod
    def has_stage_x(map):
        return AcqEngMetadata.X_UM in map

    @staticmethod
    def get_stage_x(smd):
        try:
            return smd[AcqEngMetadata.X_UM]
        except KeyError:
            raise RuntimeError("Could not get stage x")

    @staticmethod
    def set_stage_y(smd, y):
        smd[AcqEngMetadata.Y_UM] = y

    @staticmethod
    def has_stage_y(map):
        return AcqEngMetadata.Y_UM in map

    @staticmethod
    def get_stage_y(smd):
        try:
            return smd[AcqEngMetadata.Y_UM]
        except KeyError:
            raise RuntimeError("Could not get stage y")

    @staticmethod
    def set_channel_group(summary, channel_group):
        summary[AcqEngMetadata.CHANNEL_GROUP] = channel_group

    @staticmethod
    def has_channel_group(map):
        return AcqEngMetadata.CHANNEL_GROUP in map

    @staticmethod
    def get_channel_group(summary):
        try:
            return summary[AcqEngMetadata.CHANNEL_GROUP]
        except KeyError:
            raise RuntimeError("Could not find Channel Group")

    @staticmethod
    def set_core_autofocus(summary, auto_focus_device):
        summary[AcqEngMetadata.CORE_AUTOFOCUS_DEVICE] = auto_focus_device

    @staticmethod
    def has_core_autofocus(summary):
        return AcqEngMetadata.CORE_AUTOFOCUS_DEVICE in summary

    @staticmethod
    def get_core_autofocus_device(summary):
        try:
            return summary[AcqEngMetadata.CORE_AUTOFOCUS_DEVICE]
        except KeyError:
            raise ValueError("Could not find autofocus device")

    @staticmethod
    def set_core_camera(summary, camera_device):
        summary[AcqEngMetadata.CORE_CAMERA] = camera_device

    @staticmethod
    def has_core_camera(summary):
        return AcqEngMetadata.CORE_CAMERA in summary

    @staticmethod
    def get_core_camera(summary):
        try:
            return summary[AcqEngMetadata.CORE_CAMERA]
        except KeyError:
            raise ValueError("Could not get core camera")

    @staticmethod
    def set_core_galvo(summary, galvo_device):
        summary[AcqEngMetadata.CORE_GALVO] = galvo_device

    @staticmethod
    def has_core_galvo(summary):
        return AcqEngMetadata.CORE_GALVO in summary

    @staticmethod
    def get_core_galvo(summary):
        try:
            return summary[AcqEngMetadata.CORE_GALVO]
        except KeyError:
            raise ValueError("Could not get core galvo")

    @staticmethod
    def set_core_image_processor(summary, image_processor_device):
        summary[AcqEngMetadata.CORE_IMAGE_PROCESSOR] = image_processor_device

    @staticmethod
    def has_core_image_processor(summary):
        return AcqEngMetadata.CORE_IMAGE_PROCESSOR in summary

    @staticmethod
    def get_core_image_processor(summary):
        try:
            return summary[AcqEngMetadata.CORE_IMAGE_PROCESSOR]
        except KeyError:
            raise ValueError("Could not find core image processor")

    @staticmethod
    def set_core_slm(summary, slm_device):
        summary[AcqEngMetadata.CORE_SLM] = slm_device

    @staticmethod
    def has_core_slm(summary):
        return AcqEngMetadata.CORE_SLM in summary

    @staticmethod
    def get_core_slm(summary):
        try:
            return summary[AcqEngMetadata.CORE_SLM]
        except KeyError:
            raise ValueError("Could not find core slm")

    @staticmethod
    def set_core_shutter(summary, shutter_device):
        summary[AcqEngMetadata.CORE_SHUTTER] = shutter_device

    @staticmethod
    def has_core_shutter(summary):
        return AcqEngMetadata.CORE_SHUTTER in summary

    @staticmethod
    def get_core_shutter(summary):
        try:
            return summary[AcqEngMetadata.CORE_SHUTTER]
        except KeyError:
            raise ValueError("Could not find core shutter")

    @staticmethod
    def create_axes(tags):
        tags[AcqEngMetadata.AXES] = {}

    @staticmethod
    def get_axes(tags):
        try:
            axes = tags[AcqEngMetadata.AXES]
            axes_map = {}
            for key in axes:
                axes_map[key] = axes[key]
            return axes_map
        except KeyError:
            raise ValueError("Could not create axes")

    @staticmethod
    def get_axes_as_json(axes):
        try:
            axes_json = {}
            for key in axes:
                axes_json[key] = axes[key]
            return axes_json
        except KeyError:
            raise ValueError("Could not convert axes to JSON")

    @staticmethod
    def set_axis_position(tags, axis, position):
        if position is None:
            if AcqEngMetadata.has_axis(tags, axis):
                del tags[AcqEngMetadata.AXES][axis]
                return
        if not isinstance(position, (str, int, np.int64, np.int32)):
            raise ValueError("position must be String or Integer")
        tags[AcqEngMetadata.AXES][axis] = position

    @staticmethod
    def has_axis(tags, axis):
        try:
            return axis in tags[AcqEngMetadata.AXES]
        except KeyError:
            raise ValueError("Axes not present in metadata")

    @staticmethod
    def get_axis_position(tags, axis):
        try:
            return tags[AcqEngMetadata.AXES][axis]
        except KeyError:
            raise ValueError("Could not create axes")


name = "pycromanager"

from pycromanager.acquisition.java_backend_acquisitions import JavaBackendAcquisition, MagellanAcquisition, XYTiledAcquisition, ExploreAcquisition
from pycromanager.acquisition.acquisition_superclass import multi_d_acquisition_events
from pycromanager.acquisition.acq_constructor import Acquisition
from pycromanager.headless import start_headless, stop_headless
from pycromanager.mm_java_classes import Studio, Magellan
from pycromanager.core import Core
from pyjavaz import JavaObject, JavaClass, PullSocket, PushSocket
from pycromanager.acquisition.acq_eng_py.main.acq_notification import AcqNotification
from pycromanager.logging import set_logger_instance, reset_logger_instance
from ndtiff import Dataset
from ._version import __version__, version_info



"""
tests for acquisition hooks, image processors, image_saved functions, etc
"""
import numpy as np
from pycromanager import Acquisition, multi_d_acquisition_events


def test_img_process_fn(launch_mm_headless, setup_data_folder):
    events = multi_d_acquisition_events(num_time_points=3)

    def hook_fn(image, metadata):
        assert np.sum(image) > 0
        assert isinstance(metadata, dict)

        if metadata['Axes']["time"] == 1:
            image = np.zeros_like(image)

        return image, metadata

    with Acquisition(setup_data_folder, 'acq', show_display=False,
                                image_process_fn=hook_fn) as acq:
        acq.acquire(events)

    dataset = acq.get_dataset()
    try:
        data = dataset.as_array(axes=['time'])
        assert np.sum(data[0]) > 0
        assert np.sum(data[1]) == 0
        assert np.sum(data[2]) > 0
    finally:
        dataset.close()


def test_img_process_fn_no_save(launch_mm_headless):
    events = multi_d_acquisition_events(num_time_points=3)

    def hook_fn(image, metadata):
        return None

    with Acquisition(directory=None, name='acq', show_display=False,
                                image_process_fn=hook_fn) as acq:
        acq.acquire(events)
        dataset = acq.get_dataset()  # Can this be moved out of the Acquisition context?

    assert dataset is None


def test_img_process_fn_image_saved_fn_consistency(launch_mm_headless, setup_data_folder):
    def processed(image, metadata):
        processed.num_processed += 1
        return image, metadata
    processed.num_processed = 0

    def saved(_axis, _dataset):
        saved.num_saved += 1
    saved.num_saved = 0

    with Acquisition(directory=setup_data_folder, name="acq",
                     image_saved_fn=saved, image_process_fn=processed,
                     show_display=False) as acq:
        acq.acquire(multi_d_acquisition_events(num_time_points=200))

    assert(processed.num_processed == 200)
    assert(saved.num_saved == 200)

def test_event_serialize_and_deserialize(launch_mm_headless):
    """
    Test for cycle consistency of event serialization and deserialization.
    """

    events = [
        {'axes': {'channel': 'DAPI'},
         'config_group': ['Channel', 'DAPI']},
        {'axes': {'view': 0},
         'camera': 'Camera'},
        {'axes': {'z': 4},
         'z': 123.34},
        {'axes': {'time': 0},
         'exposure': 100.0},
        {'axes': {'time': 1},
         'properties': [['DeviceName', 'PropertyName', 'PropertyValue']]},
        {'axes': {'z': 1},
         'stage_positions': [['ZDeviceName', 123.45]]},
        {'axes': {'time': 2},
         'timeout': 1000},
    ]

    def hook_fn(event):
        test_event = events.pop(0)
        assert (event == test_event)
        return None  # cancel the event

    with Acquisition(show_display=False, pre_hardware_hook_fn=hook_fn) as acq:
        # copy list of events to avoid popping from original
        events_copy = [e for e in events]
        for test_event in events:
            acq.acquire(test_event)





import threading
from pycromanager.acquisition.acq_eng_py.main.acq_notification import AcqNotification

def _axes_to_key(axes):
    """ Turn axes into a hashable key """
    return None if axes is None else frozenset(axes.items())

class AcquisitionFuture:

    def __init__(self, acq, axes_or_axes_list):
        """
        :param event_or_events: a single event (dictionary) or a list of events
        """
        self._acq = acq
        self._condition = threading.Condition()
        self._notification_recieved = {}
        if isinstance(axes_or_axes_list, dict):
            axes_or_axes_list = [axes_or_axes_list]
        for axes in axes_or_axes_list:
            # single event
            # TODO maybe unify snap and sequence cause this is confusing
            self._notification_recieved[_axes_to_key(axes)] = {
                AcqNotification.Hardware.PRE_HARDWARE: False,
                AcqNotification.Hardware.POST_HARDWARE: False,

                AcqNotification.Camera.PRE_SNAP: False,
                AcqNotification.Camera.PRE_SEQUENCE_STARTED: False,
                AcqNotification.Camera.POST_EXPOSURE: False,

                AcqNotification.Image.IMAGE_SAVED: False,
            }


    def _notify(self, notification):
        """
        Called by the internal notification dispatcher in order so that it can check off that the notification was
        received. Want to store this, rather than just waiting around for it, in case the await methods are called
        after the notification has already been sent.
        """
        if notification.phase == AcqNotification.Acquisition.ACQ_EVENTS_FINISHED or \
            notification.phase == AcqNotification.Image.DATA_SINK_FINISHED:
            return # ignore for now...
        if isinstance(notification.id, list):
            keys = [_axes_to_key(ax) for ax in notification.id]
        else:
            keys = [_axes_to_key(notification.id)]
        # check if any keys are present in the notification_recieved dict
        if not any([key in self._notification_recieved.keys() for key in keys]):
            return # ignore notifications that aren't relevant to this future
        for key in keys:
            self._notification_recieved[key][notification.phase] = True
        with self._condition:
            self._condition.notify_all()

    def await_execution(self, axes, phase):
        key = _axes_to_key(axes)
        if key not in self._notification_recieved.keys() or phase not in self._notification_recieved[key].keys():
            notification = AcqNotification(None, axes, phase)
            raise ValueError("this future is not expecting a notification for: " + str(notification.to_json()))
        with self._condition:
            while not self._notification_recieved[key][phase]:
                self._condition.wait()

    def await_image_saved(self, axes, return_image=False):
        if isinstance(axes, list):
            keys = [_axes_to_key(ax) for ax in axes]
        else:
            keys = [_axes_to_key(axes)]
        for key in keys:
            with self._condition:
                while not self._notification_recieved[key][AcqNotification.Image.IMAGE_SAVED]:
                    self._condition.wait()
        if return_image:
            if isinstance(axes, list):
                return [self._acq.get_dataset().read_image(**ax) for ax in axes]
            else:
                return self._acq.get_dataset().read_image(**axes)





"""
Generic acquisition functionality used by both Python and Java backends
"""

import copy
import types
import numpy as np
from typing import Union, List, Iterable
import warnings
from abc import ABCMeta, abstractmethod
from docstring_inheritance import NumpyDocstringInheritanceMeta
import queue
import weakref
from pycromanager.acq_future import AcqNotification, AcquisitionFuture
import os
import threading
from inspect import signature


class AcqAlreadyCompleteException(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

# Subclasses inherit docstrings from abstract base class
class Meta(ABCMeta, NumpyDocstringInheritanceMeta):
  pass

class Acquisition(metaclass=Meta):

    def __init__(
            self,
            directory: str = None,
            name: str = "default_acquisition_name",
            image_process_fn: callable = None,
            event_generation_hook_fn: callable = None,
            pre_hardware_hook_fn: callable = None,
            post_hardware_hook_fn: callable = None,
            post_camera_hook_fn: callable = None,
            notification_callback_fn: callable = None,
            image_saved_fn: callable = None,
            napari_viewer=None,
            debug: int = False
    ):
        """
        Parameters
        ----------
        directory : str
            saving directory for this acquisition. If it is not supplied, the image data will be stored in RAM
            (Java backend only)
        name : str
            Name of the acquisition. This will be used to generate the folder where the data is saved.
        image_process_fn : Callable
            image processing function that will be called on each image that gets acquired.
            Can either take two arguments (image, metadata) where image is a numpy array and metadata is a dict
            containing the corresponding image metadata. Or a three argument version is accepted, which accepts (image,
            metadata, queue), where queue is a Queue object that holds upcoming acquisition events. The function
            should return either an (image, metadata) tuple or a list of such tuples
        event_generation_hook_fn : Callable
            hook function that will as soon as acquisition events are generated (before hardware sequencing optimization
            in the acquisition engine. This is useful if one wants to modify acquisition events that they didn't generate
            (e.g. those generated by a GUI application). Accepts either one argument (the current acquisition event)
            or two arguments (current event, event_queue)
        pre_hardware_hook_fn : Callable
            hook function that will be run just before the hardware is updated before acquiring
            a new image. In the case of hardware sequencing, it will be run just before a sequence of instructions are
            dispatched to the hardware. Accepts either one argument (the current acquisition event) or two arguments
            (current event, event_queue)
        post_hardware_hook_fn : Callable
            hook function that will be run just before the hardware is updated before acquiring
            a new image. In the case of hardware sequencing, it will be run just after a sequence of instructions are
            dispatched to the hardware, but before the camera sequence has been started. Accepts either one argument
            (the current acquisition event) or two arguments (current event, event_queue)
        post_camera_hook_fn : Callable
            hook function that will be run just after the camera has been triggered to snapImage or
            startSequence. A common use case for this hook is when one want to send TTL triggers to the camera from an
            external timing device that synchronizes with other hardware. Accepts either one argument (the current
            acquisition event) or two arguments (current event, event_queue)
        notification_callback_fn : Callable
            function that will be called whenever a notification is received from the acquisition engine. These
            include various stages of the control of hardware and the camera and saving of images. Notification
            callbacks will execute asynchronously with respect to the acquisition process. The supplied function
            should take a single argument, which will be an AcqNotification object. It should execute quickly,
            so as to not back up the processing of other notifications.
        image_saved_fn : Callable
            function that takes two arguments (the Axes of the image that just finished saving, and the Dataset)
            or three arguments (Axes, Dataset and the event_queue) and gets called whenever a new image is written to
            disk
        napari_viewer : napari.Viewer
            Provide a napari viewer to display acquired data in napari (https://napari.org/) rather than the built-in
            NDViewer. None by default. Data is added to the 'pycromanager acquisition' layer, which may be pre-configured by
            the user
        debug : bool
            whether to print debug messages
        show_display : bool
            If True, show the image viewer window. If False, show no viewer. (Java backend only)
        saving_queue_size : int
            The number of images to queue (in memory) while waiting to write to disk. Higher values should
            in theory allow sequence acquisitions to go faster, but requires the RAM to hold images while
            they are waiting to save (Java backend only)
        timeout :
            Timeout in ms for connecting to Java side (Java backend only)
        port :
            Allows overriding the default port for using Java backends on a different port. Use this
            after calling start_headless with the same non-default port (Java backend only)
        """
        self._debug = debug
        self._dataset = None
        self._finished = False
        self._exception = None
        self._napari_viewer = None
        self._notification_queue = queue.Queue(100)
        self._image_notification_queue = queue.Queue(100)
        self._acq_futures = []
        self._image_process_fn = image_process_fn

        pass


    def _start_notification_dispatcher(self, notification_callback_fn):
        """
        Thread that runs a function that pulls notifications from the queueand dispatches
        them to the appropriate listener
        """
        def dispatch_notifications():
            events_finished = False
            data_sink_finished = False
            while True:
                # dispatch notifications to all listeners
                notification = self._notification_queue.get()

                if AcqNotification.is_acquisition_finished_notification(notification):
                    events_finished = True
                elif AcqNotification.is_data_sink_finished_notification(notification):
                    data_sink_finished = True
                # notify acquisition futures so they can stop blocking
                for future in self._acq_futures:
                    strong_ref = future()
                    if strong_ref is not None:
                        strong_ref._notify(notification)
                # alert user-specified notification callback
                if notification_callback_fn is not None:
                    notification_callback_fn(notification)

                if events_finished and data_sink_finished:
                    break

        dispatcher_thread = threading.Thread(
            target=dispatch_notifications,
            name="NotificationDispatcherThread",
        )
        dispatcher_thread.start()
        return dispatcher_thread

    @abstractmethod
    def get_dataset(self):
        """
        Get access to the dataset backing this acquisition
        """

    @abstractmethod
    def await_completion(self):
        """
        Wait for acquisition to finish and resources to be cleaned up. If data is being written to
        disk, this will wait for the data to be written before returning.
        """

    @abstractmethod
    def get_viewer(self):
        """
        Return a reference to the current viewer, if the show_display argument
        was set to True. The returned object is either an instance of NDViewer or napari.Viewer()
        """

    def mark_finished(self):
        """
        Signal to acquisition that no more events will be added and it is time to initiate shutdown.
        This is only needed if the context manager (i.e. "with Acquisition...") is not used.
        """
        # Some acquisition types (e.g. ExploreAcquisitions) generate their own events
        # and don't send events over a port
        if self._event_queue is not None:
            # this should shut down storage and viewer as appropriate
            self._event_queue.put(None)

    def acquire(self, event_or_events: dict or list):
        """
        Submit an event or a list of events for acquisition. A single event is a python dictionary
        with a specific structure. The acquisition engine will determine if multiple events can
        be merged into a hardware sequence and executed at once without computer-hardware communication in
        between. This sequencing will only take place for events that are within a single call to acquire,
        so if you want to ensure this doesn't happen, call acquire multiple times with each event in a
        list individually.

        Parameters
        ----------
        event_or_events  : list, dict
            A single acquistion event (a dict) or a list of acquisition events

        """
        if self._acq.are_events_finished():
            raise AcqAlreadyCompleteException(
                'Cannot submit more events because this acquisition is already finished')

        if event_or_events is None:
            # manual shutdown
            self._event_queue.put(None)
            return

        _validate_acq_events(event_or_events)

        axes_or_axes_list = event_or_events['axes'] if type(event_or_events) == dict\
            else [e['axes'] for e in event_or_events]
        acq_future = AcquisitionFuture(self, axes_or_axes_list)
        self._acq_futures.append(weakref.ref(acq_future))
        # clear out old weakrefs
        self._acq_futures = [f for f in self._acq_futures if f() is not None]

        self._event_queue.put(event_or_events)
        return acq_future



    def abort(self, exception=None):
        """
        Cancel any pending events and shut down immediately

        Parameters
        ----------
        exception  : Exception
            The exception that is the reason abort is being called
        """
        # Store the exception that caused this
        if exception is not None:
            self._exception = exception

        # Clear any pending events on the python side, if applicable
        if self._event_queue is not None:
            self._event_queue.queue.clear()
            # Don't send any more events. The event sending thread should know shut itself down by
            # checking the status of the acquisition
        self._acq.abort()

    def _create_event_queue(self):
        """Create thread safe queue for events so they can be passed from multiple processes"""
        self._event_queue = queue.Queue()

    def _call_image_process_fn(self, image, metadata):
        params = signature(self._process_fn).parameters
        processed = None
        if len(params) == 2 or len(params) == 3:
            try:
                if len(params) == 2:
                    processed = self._process_fn(image, metadata)
                elif len(params) == 3:
                    processed = self._process_fn(image, metadata, self._event_queue)
            except Exception as e:
                self.abort(Exception("exception in image processor: {}".format(e)))

        else:
            self.abort(Exception(
                "Incorrect number of arguments for image processing function, must be 2 or 3"
            ))
        return processed

    ########  Context manager (i.e. "with Acquisition...") ###########
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.mark_finished()
        # now wait on it to finish
        self.await_completion()


def _validate_acq_events(events: dict or list):
    """
    Validate if supplied events are a dictionary or a list of dictionaries
    that contain valid events. Throw an exception if not

    Parameters
    ----------
    events : dict or list

    """
    if isinstance(events, dict):
        _validate_acq_dict(events)
    elif isinstance(events, list):
        if len(events) == 0:
            raise Exception('events list cannot be empty')
        for event in events:
            if isinstance(event, dict):
                _validate_acq_dict(event)
            else:
                raise Exception('events must be a dictionary or a list of dictionaries')
    else:
        raise Exception('events must be a dictionary or a list of dictionaries')

def _validate_acq_dict(event: dict):
    """
    Validate event dictionary, and raise an exception or supply a warning and fix it if something is incorrect

    Parameters
    ----------
    event : dict

    """
    if 'axes' not in event.keys():
        raise Exception('event dictionary must contain an \'axes\' key. This event will be ignored')
    if 'row' in event.keys():
        warnings.warn('adding \'row\' as a top level key in the event dictionary is deprecated and will be disallowed in '
                      'a future version. Instead, add \'row\' as a key in the \'axes\' dictionary')
        event['axes']['row'] = event['row']
    if 'col' in event.keys():
        warnings.warn('adding \'col\' as a top level key in the event dictionary is deprecated and will be disallowed in '
                      'a future version. Instead, add \'column\' as a key in the \'axes\' dictionary')
        event['axes']['column'] = event['col']

    # TODO check for the validity of other acquisition event fields, and make sure that there aren't unexpected
    #   other fields, to help users catch simple errors


def multi_d_acquisition_events(
    num_time_points: int=None,
    time_interval_s: Union[float, List[float]]=0,
    z_start: float=None,
    z_end: float=None,
    z_step: float=None,
    channel_group: str=None,
    channels: list=None,
    channel_exposures_ms: list=None,
    xy_positions: Iterable=None,
    xyz_positions: Iterable=None,
    position_labels: List[str]=None,
    order: str="tpcz",
):
    """Convenience function for generating the events of a typical multi-dimensional acquisition (i.e. an
    acquisition with some combination of multiple timepoints, channels, z-slices, or xy positions)

    Parameters
    ----------
    num_time_points : int
        How many time points if it is a timelapse (Default value = None)
    time_interval_s : float or list of floats
        the minimum interval between consecutive time points in seconds. If set to 0, the 
        acquisition will go as fast as possible. If a list is provided, its length should 
        be equal to 'num_time_points'. Elements in the list are assumed to be the intervals
        between consecutive timepoints in the timelapse. First element in the list indicates
        delay before capturing the first image (Default value = 0)
    z_start : float
        z-stack starting position, in µm. If xyz_positions is given z_start is relative
        to the points' z position. (Default value = None)
    z_end : float
        z-stack ending position, in µm. If xyz_positions is given z_start is
        relative to the points' z position. (Default value = None)
    z_step : float
        step size of z-stack, in µm (Default value = None)
    channel_group : str
        name of the channel group (which should correspond to a config group in micro-manager) (Default value = None)
    channels : list of strings
        list of channel names, which correspond to possible settings of the config group
        (e.g. ['DAPI', 'FITC']) (Default value = None)
    channel_exposures_ms : list of floats or ints
        list of camera exposure times corresponding to each channel. The length of this list
        should be the same as the the length of the list of channels (Default value = None)
    xy_positions : iterable
        An array of shape (N, 2) containing N (X, Y) stage coordinates. (Default value = None)
    xyz_positions : iterable
        An array of shape (N, 3) containing N (X, Y, Z) stage coordinates. (Default value = None).
        If passed then z_start, z_end, and z_step will be relative to the z_position in xyz_positions. (Default value = None)
    position_labels : iterable
        An array of length N containing position labels for each of the XY stage positions. (Default value = None)
    order : str
        string that specifies the order of different dimensions. Must have some ordering of the letters
        c, t, p, and z. For example, 'tcz' would run a timelapse where z stacks would be acquired at each channel in
        series. 'pt' would move to different xy stage positions and run a complete timelapse at each one before moving
        to the next (Default value = 'tpcz')

    Returns
    -------
    events : dict
    """
    if xy_positions is not None and xyz_positions is not None:
        raise ValueError(
            "xyz_positions and xy_positions are incompatible arguments that cannot be passed together"
        )
    order = order.lower()
    if "p" in order and "z" in order and order.index("p") > order.index("z"):
        raise ValueError(
            "This function requires that the xy position come earlier in the order than z"
        )
    if isinstance(time_interval_s, list):
        if len(time_interval_s) != num_time_points:
            raise ValueError(
                "Length of time interval list should be equal to num_time_points"
            )
    if position_labels is not None:
        if xy_positions is not None and len(xy_positions) != len(position_labels):
            raise ValueError("xy_positions and position_labels must be of equal length")
        if xyz_positions is not None and len(xyz_positions) != len(position_labels):
            raise ValueError("xyz_positions and position_labels must be of equal length")
    
    # If any of z_start, z_step, z_end are provided, then they should all be provided
    # Here we can't use `all` as some of the values of z_start, z_step, z_end
    # may be zero and all((0,)) = False
    has_zsteps = False
    if any([z_start, z_step, z_end]):
        if not None in [z_start, z_step, z_end]:
            has_zsteps = True
        else:
            raise ValueError('All of z_start, z_step, and z_end must be provided')

    z_positions = None
    if xy_positions is not None:
        xy_positions = np.asarray(xy_positions)
        z_positions = None
    elif xyz_positions is not None:
        xyz_positions = np.asarray(xyz_positions)
        xy_positions = xyz_positions[:, :2]
        z_positions = xyz_positions[:, 2][:, None]

    if has_zsteps:
        z_rel = np.arange(z_start, z_end + z_step, z_step)
        if z_positions is None:
            z_positions = z_rel
            if xy_positions is not None:
                z_positions = np.broadcast_to(
                    z_positions, (xy_positions.shape[0], z_positions.shape[0])
                )
        else:
            pos = []
            for z in z_positions:
                pos.append(z + z_rel)
            z_positions = np.asarray(pos)

    if position_labels is None and xy_positions is not None:
        position_labels = list(range(len(xy_positions)))

    def generate_events(event, order):
        if len(order) == 0:
            yield event
            return
        elif order[0] == "t" and num_time_points is not None and num_time_points > 0:
            time_indices = np.arange(num_time_points)
            if isinstance(time_interval_s, list):
                absolute_start_times = np.cumsum(time_interval_s)
            for time_index in time_indices:
                new_event = copy.deepcopy(event)
                new_event["axes"]["time"] = time_index
                if isinstance(time_interval_s, list):
                    new_event["min_start_time"] = absolute_start_times[time_index]
                else:
                    if time_interval_s != 0:
                        new_event["min_start_time"] = time_index * time_interval_s
                yield generate_events(new_event, order[1:])
        elif order[0] == "z" and z_positions is not None:
            if "axes" in event and "position" in event["axes"]:
                pos_idx = position_labels.index(event["axes"]["position"])
                zs = z_positions[pos_idx]
            else:
                zs = z_positions

            for z_index, z in enumerate(zs):
                new_event = copy.deepcopy(event)
                new_event["axes"]["z"] = z_index
                new_event["z"] = z
                yield generate_events(new_event, order[1:])
        elif order[0] == "p" and xy_positions is not None:
            for p_label, xy in zip(position_labels, xy_positions):
                new_event = copy.deepcopy(event)
                new_event["axes"]["position"] = p_label
                new_event["x"] = xy[0]
                new_event["y"] = xy[1]
                yield generate_events(new_event, order[1:])
        elif order[0] == "c" and channel_group is not None and channels is not None:
            for i in range(len(channels)):
                new_event = copy.deepcopy(event)
                new_event["config_group"] = [channel_group,  channels[i]]
                new_event["axes"]["channel"] = channels[i]
                if channel_exposures_ms is not None:
                    new_event["exposure"] = channel_exposures_ms[i]
                yield generate_events(new_event, order[1:])
        else:
            # this axis appears to be missing
            yield generate_events(event, order[1:])

    # collect all events into a single list
    base_event = {"axes": {}}
    events = []

    def appender(next):
        """

        Parameters
        ----------
        next :


        Returns
        -------

        """
        if isinstance(next, types.GeneratorType):
            for n in next:
                appender(n)
        else:
            events.append(next)

    appender(generate_events(base_event, order))
    return events





from docstring_inheritance import NumpyDocstringInheritanceMeta
from pycromanager.acquisition.acq_eng_py.main.AcqEngPy_Acquisition import Acquisition as pymmcore_Acquisition
from pycromanager.acquisition.RAMStorage import RAMDataStorage
from pycromanager.acquisition.acquisition_superclass import _validate_acq_events, Acquisition
from pycromanager.acquisition.acq_eng_py.main.acquisition_event import AcquisitionEvent
from pycromanager.acq_future import AcqNotification
import threading
from inspect import signature


class PythonBackendAcquisition(Acquisition, metaclass=NumpyDocstringInheritanceMeta):
    """
    Pycro-Manager acquisition that uses a Python runtime backend. Unlike the Java backend,
    Python-backed acquisitions currently do not automatically write data to disk. Instead, by default,
    they store data in RAM which can be queried with the Dataset class. If instead you want to
    implement your own data storage, you can pass an image_process_fn which diverts the data to
    a custom endpoint.
    """

    def __init__(
        self,
        name: str='default_acq_name',
        image_process_fn: callable=None,
        event_generation_hook_fn: callable = None,
        pre_hardware_hook_fn: callable=None,
        post_hardware_hook_fn: callable=None,
        post_camera_hook_fn: callable=None,
        notification_callback_fn: callable=None,
        napari_viewer=None,
        image_saved_fn: callable=None,
        debug: int=False,
        # Specificly so the directory arg can be absorbed and ignored without error,
        **kwargs
    ):
        # Get a dict of all named argument values (or default values when nothing provided)
        arg_names = [k for k in signature(PythonBackendAcquisition.__init__).parameters.keys() if k != 'self']
        l = locals()
        named_args = {arg_name: (l[arg_name] if arg_name in l else
                                     dict(signature(PythonBackendAcquisition.__init__).parameters.items())[arg_name].default)
                                     for arg_name in arg_names }
        if 'kwargs' in named_args:
            if 'directory' in named_args['kwargs'] and named_args['kwargs']['directory'] is not None:
                raise Exception('The directory argument is not supported in Python backend acquisitions')
            del named_args['kwargs']
        super().__init__(**named_args)
        self._dataset = RAMDataStorage()
        self._finished = False
        self._notifications_finished = False
        self._create_event_queue()

        self._process_fn = image_process_fn
        self._image_processor = ImageProcessor(self) if image_process_fn is not None else None


        # create a thread that submits events
        # events can be added to the queue through image processors, hooks, or the acquire method
        def submit_events():
            while True:
                event_or_events = self._event_queue.get()
                if event_or_events is None:
                    self._acq.finish()
                    self._acq.block_until_events_finished()
                    break
                _validate_acq_events(event_or_events)
                if isinstance(event_or_events, dict):
                    event_or_events = [event_or_events]
                # convert to objects
                event_or_events = [AcquisitionEvent.from_json(event, self._acq) for event in event_or_events]
                self._acq.submit_event_iterator(iter(event_or_events))
        self._event_thread = threading.Thread(target=submit_events)
        self._event_thread.start()

        self._acq = pymmcore_Acquisition(self._dataset)

        # receive notifications from the acquisition engine. Unlike the java_backend analog
        # of this, the python backend does not have a separate thread for notifications because
        # it can just use the one in AcqEngPy
        def post_notification(notification):
            self._notification_queue.put(notification)
            # these are processed seperately to handle image saved callback
            if AcqNotification.is_image_saved_notification(notification):
                self._image_notification_queue.put(notification)

        self._acq.add_acq_notification_listener(NotificationListener(post_notification))

        self._notification_dispatch_thread = self._start_notification_dispatcher(notification_callback_fn)

        # add hooks and image processor
        if pre_hardware_hook_fn is not None:
            self._acq.add_hook(AcquisitionHook(pre_hardware_hook_fn),self._acq.BEFORE_HARDWARE_HOOK)
        if post_hardware_hook_fn is not None:
            self._acq.add_hook(AcquisitionHook(post_hardware_hook_fn),self._acq.AFTER_HARDWARE_HOOK)
        if post_camera_hook_fn is not None:
            self._acq.add_hook(AcquisitionHook(post_camera_hook_fn),self._acq.AFTER_CAMERA_HOOK)
        if event_generation_hook_fn is not None:
            self._acq.add_hook(AcquisitionHook(event_generation_hook_fn),self._acq.EVENT_GENERATION_HOOK)
        if self._image_processor is not None:
            self._acq.add_image_processor(self._image_processor)


        if napari_viewer is not None:
            # using napari viewer
            try:
                import napari
            except:
                raise Exception('Napari must be installed in order to use this feature')
            from pycromanager.napari_util import start_napari_signalling
            assert isinstance(napari_viewer, napari.Viewer), 'napari_viewer must be an instance of napari.Viewer'
            self._napari_viewer = napari_viewer
            start_napari_signalling(self._napari_viewer, self.get_dataset())


    ########  Public API ###########
    def get_dataset(self):
        return self._dataset

    def await_completion(self):
        """Wait for acquisition to finish and resources to be cleaned up"""
        while not self._acq.are_events_finished() or (
                self._acq.get_data_sink() is not None and not self._acq.get_data_sink().is_finished()):
            self._check_for_exceptions()
            self._acq.block_until_events_finished(0.05)
            if self._acq.get_data_sink() is not None:
                self._acq.get_data_sink().block_until_finished(0.05)
            self._check_for_exceptions()
        self._event_thread.join()
        self._notification_dispatch_thread.join()

        self._acq = None
        self._finished = True

    def get_viewer(self):
        """
        Return a reference to the current viewer, if the show_display argument
        was set to True. The returned object is either an instance of NDViewer or napari.Viewer()
        """
        return self._napari_viewer

    ########  Context manager (i.e. "with Acquisition...") ###########
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.mark_finished()
        # now wait on it to finish
        self.await_completion()

    def _check_for_exceptions(self):
        """
        Check for exceptions on the python side (i.e. hooks and processors)
        or on the Java side (i.e. hardware control)
        """
        # these will throw exceptions
        self._acq.check_for_exceptions()
        if self._exception is not None:
            raise self._exception

    def _are_acquisition_notifications_finished(self):
        """
        Called by the storage to check if all notifications have been processed
        """
        return self._notifications_finished

class ImageProcessor:
    """
    This is the equivalent of RemoteImageProcessor in the Java version.
    It runs its own thread, polls the input queue for images, calls
    the process function, and puts the result in the output queue.
    """


    def __init__(self, pycromanager_acq):
        self._pycromanager_acq = pycromanager_acq

    def set_acq_and_queues(self, acq, input, output):
        self.input_queue = input
        self.output_queue = output
        self._acq = acq
        self._process_thread = threading.Thread(target=self._process)
        self._process_thread.start()

    def _process(self):
        while True:
            # wait for an image to arrive
            tagged_image = self.input_queue.get()
            if tagged_image.tags is None and tagged_image.pix is None:
                # this is a signal to stop
                self.output_queue.put(tagged_image)
                break
            process_fn_result = self._pycromanager_acq._call_image_process_fn(tagged_image.tags, tagged_image.pix)
            if process_fn_result is not None:
                self.output_queue.put(process_fn_result)
            # otherwise the image processor intercepted the image and nothing to do here

class AcquisitionHook:
    """
    Lightweight wrapper to convert function pointers to AcqEng hooks
    """

    def __init__(self, hook_fn):
        self._hook_fn = hook_fn

    def run(self, event):
        self._hook_fn(event)

    def close(self):
        pass # nothing to do here

class NotificationListener:
    """
    Lightweight wrapper to convert function pointers to AcqEng notification listeners
    """

    def __init__(self, notification_fn):
        self._notification_fn = notification_fn

    def post_notification(self, notification):
        self._notification_fn(notification)


# A class for holding data in RAM

from pycromanager.acquisition.acq_eng_py.main.acq_eng_metadata import AcqEngMetadata
import numpy as np
from sortedcontainers import SortedSet
import threading


class RAMDataStorage:
    """
    A class for holding data in RAM
    Implements the methods needed to be a DataSink for AcqEngPy
    """

    def __init__(self):
        self.finished = False
        self.images = {}
        self.image_metadata = {}
        self.axes = {}
        self.finished_event = threading.Event()

    def initialize(self, acq, summary_metadata: dict):
        self.summary_metadata = summary_metadata

    def block_until_finished(self, timeout=None):
        self.finished_event.wait(timeout=timeout)

    def finish(self):
        self.finished = True

    def is_finished(self) -> bool:
        return self.finished

    def put_image(self, tagged_image):
        self.bytes_per_pixel = tagged_image.pix.dtype.itemsize
        self.dtype = tagged_image.pix.dtype
        tags = tagged_image.tags
        axes = AcqEngMetadata.get_axes(tags)
        key = frozenset(axes.items())
        self.images[key] = tagged_image.pix
        self.image_metadata[key] = tags
        for axis in axes.keys():
            if axis not in self.axes:
                self.axes[axis] = SortedSet()
            self.axes[axis].add(axes[axis])
        self._new_image_arrived = True

    def anything_acquired(self) -> bool:
        return self.images != {}

    def has_image(self, channel: int or str, z: int, position: int,
                          time: int, row: int, column: int, **kwargs):
        axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)
        key = frozenset(axes.items())
        return key in self.images.keys()

    def read_image(self, channel=None, z=None, time=None, position=None, row=None, column=None, **kwargs):
        axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)
        key = frozenset(axes.items())
        if key not in self.index:
            raise Exception("image with keys {} not present in data set".format(key))
        return self.images[key]

    def read_metadata(self, channel=None, z=None, time=None, position=None, row=None, column=None, **kwargs):
        axes = self._consolidate_axes(channel, z, position, time, row, column, **kwargs)
        key = frozenset(axes.items())
        if key not in self.index:
            raise Exception("image with keys {} not present in data set".format(key))
        return self.image_metadata[key]

    def _consolidate_axes(self, channel: int or str, z: int, position: int,
                          time: int, row: int, column: int, **kwargs):
        """
        Pack axes into a convenient format
        """
        axis_positions = {'channel': channel, 'z': z, 'position': position,
                          'time': time, 'row': row, 'column': column, **kwargs}
        # ignore ones that are None
        axis_positions = {n: axis_positions[n] for n in axis_positions.keys() if axis_positions[n] is not None}
        for axis_name in axis_positions.keys():
            # convert any string-valued axes passed as ints into strings
            if self.axes_types[axis_name] == str and type(axis_positions[axis_name]) == int:
                axis_positions[axis_name] = self._string_axes_values[axis_name][axis_positions[axis_name]]

        return axis_positions

    def has_new_image(self):
        """
        For datasets currently being acquired, check whether a new image has arrived since this function
        was last called, so that a viewer displaying the data can be updated.
        """
        # pass through to full resolution, since only this is monitored in current implementation
        if not hasattr(self, '_new_image_arrived'):
            return False # pre-initilization
        new = self._new_image_arrived
        self._new_image_arrived = False
        return new

    def as_array(self, axes=None, **kwargs):
        """
        Read all data image data as one big numpy array with last two axes as y, x and preceeding axes depending on data.
        If the data doesn't fully fill out the array (e.g. not every z-slice collected at every time point), zeros will
        be added automatically.

        This function is modeled of the same one in the NDTiff library, but it uses numpy arrays instead of dask arrays
        because the data is already in RAM

        Parameters
        ----------
        axes : list
            list of axes names over which to iterate and merge into a stacked array. The order of axes supplied in this
            list will be the order of the axes of the returned dask array. If None, all axes will be used in PTCZYX order.

        **kwargs :
            names and integer positions of axes on which to slice data
        """
        if axes is None:
            axes = self.axes.keys()

        empty_image = np.zeros_like(list(self.images.values())[0])
        indices = [np.array(self.axes[axis_name]) for axis_name in list(axes)]
        gridded = np.meshgrid(*indices, indexing='ij')
        result = np.stack(gridded, axis=-1)
        flattened = result.reshape((-1, result.shape[-1]))
        images = []
        for coord in flattened:
            images_key = {key: coord[i] for i, key in enumerate(axes)}
            key = frozenset(images_key.items())
            if key in self.images.keys():
                images.append(self.images[key])
            else:
                images.append(empty_image)
        # reshape to Num axes + image size dimensions
        cube = np.array(images).reshape(tuple(len(i) for i in indices) + empty_image.shape)
        return cube






"""
Tests for notification API and asynchronous acquisition API
"""
from pycromanager import multi_d_acquisition_events, Acquisition, AcqNotification
import time
import numpy as np

# TODO: add tests for timing of blocking until different parts of the hardware sequence
# def test_async_images_read(launch_mm_headless, setup_data_folder):
#     start = time.time()
#     events = multi_d_acquisition_events(num_time_points=10, time_interval_s=0.5)
#     with Acquisition(directory=setup_data_folder, show_display=False) as acq:
#         future = acq.acquire(events)
#
#         future.await_execution({'time': 5}, AcqNotification.Hardware.POST_HARDWARE)



def test_async_image_read(launch_mm_headless, setup_data_folder):
    events = multi_d_acquisition_events(num_time_points=10, time_interval_s=0.5)
    with Acquisition(directory=setup_data_folder, show_display=False) as acq:
        future = acq.acquire(events)
        image = future.await_image_saved({'time': 5}, return_image=True)
        assert np.all(image == acq.get_dataset().read_image(time=5))

def test_async_image_read_sequence(launch_mm_headless, setup_data_folder):
    events = multi_d_acquisition_events(num_time_points=10, time_interval_s=0)
    with Acquisition(directory=setup_data_folder, show_display=False) as acq:
        future = acq.acquire(events)
        image = future.await_image_saved({'time': 5}, return_image=True)
        assert np.all(image == acq.get_dataset().read_image(time=5))

def test_async_images_read(launch_mm_headless, setup_data_folder):
    events = multi_d_acquisition_events(num_time_points=10, time_interval_s=0.5)
    with Acquisition(directory=setup_data_folder, show_display=False) as acq:
        future = acq.acquire(events)
        images = future.await_image_saved([{'time': 7}, {'time': 8}, {'time': 9}], return_image=True)
        assert (len(images) == 3)

    # Make sure the returned images were the correct ones
    on_disk = [acq.get_dataset().read_image(time=t) for t in [7, 8, 9]]
    assert all([np.all(on_disk[i] == images[i]) for i in range(3)])

def test_async_images_read_sequence(launch_mm_headless, setup_data_folder):
    events = multi_d_acquisition_events(num_time_points=10, time_interval_s=0)
    with Acquisition(directory=setup_data_folder, show_display=False) as acq:
        future = acq.acquire(events)
        images = future.await_image_saved([{'time': 7}, {'time': 8}, {'time': 9}], return_image=True)
        assert (len(images) == 3)

    # Make sure the returned images were the correct ones
    on_disk = [acq.get_dataset().read_image(time=t) for t in [7, 8, 9]]
    assert all([np.all(on_disk[i] == images[i]) for i in range(3)])




import logging

logging.basicConfig(level=logging.DEBUG)
main_logger = _base_logger = logging.getLogger("pycromanager")

def set_logger_instance(logger_name: str) -> None:
    """ Replaces the default logger with a custom logger
    
    logger (str)
        name of new logger to use (i.e. logger.name)
    """
    global main_logger
    main_logger = logging.getLogger(logger_name)

def reset_logger_instance() -> None:
    """ Resets the logger to the default logger """
    set_logger_instance(_base_logger.name)


import json
import queue
import traceback
import threading

from pycromanager.acquisition.acq_eng_py.main.acq_eng_metadata import AcqEngMetadata
from pycromanager.acquisition.acq_eng_py.internal.engine import Engine
from pycromanager.acquisition.acq_eng_py.main.acq_notification import AcqNotification
from pycromanager.acquisition.acq_eng_py.internal.notification_handler import NotificationHandler



class Acquisition():

    EVENT_GENERATION_HOOK = 0
    # This hook runs before changes to the hardware (corresponding to the instructions in the
    # event) are made
    BEFORE_HARDWARE_HOOK = 1
    # This hook runs after changes to the hardware took place, but before camera exposure
    # (either a snap or a sequence) is started
    AFTER_HARDWARE_HOOK = 2
    # Hook runs after the camera sequence acquisition has started. This can be used for
    # external triggering of the camera
    AFTER_CAMERA_HOOK = 3
    # Hook runs after the camera exposure ended (when possible, before readout of the camera
    # and availability of the images in memory).
    AFTER_EXPOSURE_HOOK = 4

    IMAGE_QUEUE_SIZE = 30

    def __init__(self, sink, summary_metadata_processor=None, initialize=True):
        self.xy_stage_ = None
        self.events_finished_ = threading.Event()
        self.abort_requested_ = threading.Event()
        self.start_time_ms_ = -1
        self.paused_ = False
        self.event_generation_hooks_ = []
        self.before_hardware_hooks_ = []
        self.after_hardware_hooks_ = []
        self.after_camera_hooks_ = []
        self.after_exposure_hooks_ = []
        self.image_processors_ = []
        self.first_dequeue_ = queue.Queue(maxsize=self.IMAGE_QUEUE_SIZE)
        self.processor_output_queues_ = {}
        self.debug_mode_ = False
        self.abort_exception_ = None
        self.image_metadata_processor_ = None
        self.notification_handler_ = NotificationHandler()
        self.started_ = False
        self.core_ = Engine.get_core()
        self.summary_metadata_processor_ = summary_metadata_processor
        self.data_sink_ = sink
        if initialize:
            self.initialize()

    def post_notification(self, notification):
        self.notification_handler_.post_notification(notification)

    def add_acq_notification_listener(self, post_notification_fn):
        self.notification_handler_.add_listener(post_notification_fn)

    def get_data_sink(self):
        return self.data_sink_

    def set_debug_mode(self, debug):
        self.debug_mode_ = debug

    def is_debug_mode(self):
        return self.debug_mode_

    def is_abort_requested(self):
        return self.abort_requested_.is_set()

    def abort(self, e=None):
        if e:
            self.abort_exception_ = e
        if self.abort_requested_.is_set():
            return
        self.abort_requested_.set()
        if self.is_paused():
            self.set_paused(False)
        Engine.get_instance().finish_acquisition(self)

    def check_for_exceptions(self):
        if self.abort_exception_:
            raise self.abort_exception_

    def add_to_summary_metadata(self, summary_metadata):
        if self.summary_metadata_processor_:
            self.summary_metadata_processor_(summary_metadata)

    def add_to_image_metadata(self, tags):
        if self.image_metadata_processor_:
            self.image_metadata_processor_(tags)

    def add_tags_to_tagged_image(self, tags, more_tags):
        if not more_tags:
            return
        more_tags_object = json.loads(json.dumps(more_tags))
        tags['AcqEngMetadata.TAGS'] = more_tags_object

    def submit_event_iterator(self, evt):
        if not self.started_:
            self.start()
        return Engine.get_instance().submit_event_iterator(evt)

    def start_saving_thread(self):
        def saving_thread(acq):
            try:
                while True:
                    if acq.debug_mode_:
                        acq.core_.log_message(f"Image queue size: {len(acq.first_dequeue_)}")
                    if not acq.image_processors_:
                        if acq.debug_mode_:
                            acq.core_.log_message("waiting for image to save")
                        img = acq.first_dequeue_.get()
                        if acq.debug_mode_:
                            acq.core_.log_message("got image to save")
                        if img.tags is None and img.pix is None:
                            break
                        acq.save_image(img)
                    else:
                        img = acq.processor_output_queues_[acq.image_processors_[-1]].get()
                        if acq.data_sink_:
                            if acq.debug_mode_:
                                acq.core_.log_message("Saving image")
                            if not img.pix and not img.tags:
                                break
                            acq.save_image(img)
                            if acq.debug_mode_:
                                acq.core_.log_message("Finished saving image")
            except Exception as ex:
                traceback.print_exc()
                acq.abort(ex)
            finally:
                acq.save_image(acq.core_.TaggedImage(None, None))

        threading.Thread(target=saving_thread, args=(self,)).start()

    def add_image_processor(self, p):
        if self.started_:
            raise RuntimeError("Cannot add processor after acquisition started")
        self.image_processors_.append(p)
        self.processor_output_queues_[p] = queue.Queue(maxsize=self.IMAGE_QUEUE_SIZE)
        if len(self.image_processors_) == 1:
            p.set_acq_and_queues(self, self.first_dequeue_, self.processor_output_queues_[p])
        else:
            p.set_acq_and_queues(self, self.processor_output_queues_[self.image_processors_[-2]],
                                   self.processor_output_queues_[self.image_processors_[-1]])

    def add_hook(self, h, type_):
        if self.started_:
            raise RuntimeError("Cannot add hook after acquisition started")
        if type_ == self.EVENT_GENERATION_HOOK:
            self.event_generation_hooks_.append(h)
        elif type_ == self.BEFORE_HARDWARE_HOOK:
            self.before_hardware_hooks_.append(h)
        elif type_ == self.AFTER_HARDWARE_HOOK:
            self.after_hardware_hooks_.append(h)
        elif type_ == self.AFTER_CAMERA_HOOK:
            self.after_camera_hooks_.append(h)
        elif type_ == self.AFTER_EXPOSURE_HOOK:
            self.after_exposure_hooks_.append(h)

    def initialize(self):
        if self.core_:
            summary_metadata = AcqEngMetadata.make_summary_metadata(self.core_, self)
            self.add_to_summary_metadata(summary_metadata)
            try:
                self.summary_metadata_ = summary_metadata
            except json.JSONDecodeError:
                print("Couldn't copy summary metadata")
            if self.data_sink_:
                self.data_sink_.initialize(self, summary_metadata)

    def start(self):
        if self.data_sink_:
            self.start_saving_thread()
        self.post_notification(AcqNotification.create_acq_started_notification())
        self.started_ = True

    def save_image(self, image):
        if image.tags is None and image.pix is None:
            self.data_sink_.finish()
            self.post_notification(AcqNotification.create_data_sink_finished_notification())
        else:
            self.data_sink_.put_image(image)
            axes = AcqEngMetadata.get_axes(image.tags)
            self.post_notification(AcqNotification.create_image_saved_notification(axes))

    def get_start_time_ms(self):
        return self.start_time_ms_

    def set_start_time_ms(self, time):
        self.start_time_ms_ = time

    def is_paused(self):
        return self.paused_

    def is_started(self):
        return self.started_

    def set_paused(self, pause):
        self.paused_ = pause

    def get_summary_metadata(self):
        return self.summary_metadata_

    def anything_acquired(self):
        return not self.data_sink_ or self.data_sink_.anything_acquired()

    def add_image_metadata_processor(self, processor):
        if not self.image_metadata_processor_:
            self.image_metadata_processor_ = processor
        else:
            raise RuntimeError("Multiple metadata processors not supported")

    def get_event_generation_hooks(self):
        return self.event_generation_hooks_

    def get_before_hardware_hooks(self):
        return self.before_hardware_hooks_

    def get_after_hardware_hooks(self):
        return self.after_hardware_hooks_

    def get_after_camera_hooks(self):
        return self.after_camera_hooks_

    def get_after_exposure_hooks(self):
        return self.after_exposure_hooks_

    def add_to_output(self, ti):
        try:
            if ti.tags is None and ti.pix is None:
                self.events_finished_.set()
            self.first_dequeue_.put(ti)
        except Exception as ex:
            raise RuntimeError(ex)

    def finish(self):
        Engine.get_instance().finish_acquisition(self)

    def are_events_finished(self):
        return self.events_finished_.is_set()

    def block_until_events_finished(self, timeout=None):
        """Blocks until all events have been processed."""
        self.events_finished_.wait(timeout)

    def block_unless_aborted(self, timeout_ms=None):
        """Blocks until acquisition is aborted."""
        self.abort_requested_.wait(timeout_ms / 1000)


    def get_image_transfer_queue_size(self):
        return self.IMAGE_QUEUE_SIZE

    def get_image_transfer_queue_count(self):
        return len(self.first_dequeue_)





from collections import namedtuple
import json
from pycromanager.acquisition.acq_eng_py.main.acq_eng_metadata import AcqEngMetadata

class AcquisitionEvent:
    class SpecialFlag:
        ACQUISITION_FINISHED = "AcqusitionFinished"
        ACQUISITION_SEQUENCE_END = "AcqusitionSequenceEnd"

    def __init__(self, acq, sequence=None):
        self.acquisition_ = acq
        self.axisPositions_ = {}
        self.camera_ = None
        self.timeout_ms_ = None
        self.configGroup_ = None
        self.configPreset_ = None
        self.exposure_ = None
        self.miniumumStartTime_ms_ = None
        self.zPosition_ = None
        self.xPosition_ = None
        self.yPosition_ = None
        self.stageCoordinates_ = {}
        self.stageDeviceNamesToAxisNames_ = {}
        self.tags_ = {}
        self.acquireImage_ = None
        self.slmImage_ = None
        self.properties_ = set()
        self.sequence_ = None
        self.xySequenced_ = False
        self.zSequenced_ = False
        self.exposureSequenced_ = False
        self.configGroupSequenced_ = False
        self.specialFlag_ = None

        if sequence:
            self.acquisition_ = sequence[0].acquisition_
            self.miniumumStartTime_ms_ = sequence[0].miniumumStartTime_ms_
            self.sequence_ = list(sequence)
            zPosSet = set()
            xPosSet = set()
            yPosSet = set()
            exposureSet = set()
            configSet = set()
            for event in self.sequence_:
                if event.zPosition_:
                    zPosSet.add(event.get_z_position())
                if event.xPosition_:
                    xPosSet.add(event.get_x_position())
                if event.yPosition_:
                    yPosSet.add(event.get_y_position())
                if event.exposure_:
                    exposureSet.add(event.get_exposure())
                if event.configPreset_:
                    configSet.add(event.get_config_preset())
            self.exposureSequenced_ = len(exposureSet) > 1
            self.configGroupSequenced_ = len(configSet) > 1
            self.xySequenced_ = len(xPosSet) > 1 and len(yPosSet) > 1
            self.zSequenced_ = len(zPosSet) > 1
            if sequence[0].exposure_ and not self.exposureSequenced_:
                self.exposure_ = sequence[0].exposure_


    def copy(self):
        e = AcquisitionEvent(self.acquisition_)
        e.axisPositions_ = self.axisPositions_.copy()
        e.configPreset_ = self.configPreset_
        e.configGroup_ = self.configGroup_
        e.stageCoordinates_ = self.stageCoordinates_.copy()
        e.stageDeviceNamesToAxisNames_ = self.stageDeviceNamesToAxisNames_.copy()
        e.xPosition_ = self.xPosition_
        e.yPosition_ = self.yPosition_
        e.zPosition_ = self.zPosition_
        e.miniumumStartTime_ms_ = self.miniumumStartTime_ms_
        e.slmImage_ = self.slmImage_
        e.acquireImage_ = self.acquireImage_
        e.properties_ = set(self.properties_)
        e.camera_ = self.camera_
        e.timeout_ms_ = self.timeout_ms_
        e.setTags(self.tags_)  # Assuming setTags is a method in the class
        return e

    @staticmethod
    def event_to_json(e):
        data = {}

        if e.isAcquisitionFinishedEvent():
            data["special"] = "acquisition-end"
            return json.dumps(data)
        elif e.isAcquisitionSequenceEndEvent():
            data["special"] = "sequence-end"
            return json.dumps(data)

        if e.miniumumStartTime_ms_:
            data["min_start_time"] = e.miniumumStartTime_ms_ / 1000

        if e.hasConfigGroup():
            data["config_group"] = [e.configGroup_, e.configPreset_]

        if e.exposure_:
            data["exposure"] = e.exposure_

        if e.slmImage_:
            data["slm_pattern"] = e.slmImage_

        if e.timeout_ms_:
            data["timeout_ms"] = e.timeout_ms_

        axes = {axis: e.axisPositions_[axis] for axis in e.axisPositions_}
        if axes:
            data["axes"] = axes

        stage_positions = [[stageDevice, e.getStageSingleAxisStagePosition(stageDevice)] for stageDevice in e.getStageDeviceNames()]
        if stage_positions:
            data["stage_positions"] = stage_positions

        if e.zPosition_:
            data["z"] = e.zPosition_

        if e.xPosition_:
            data["x"] = e.xPosition_

        if e.yPosition_:
            data["y"] = e.yPosition_

        if e.camera_:
            data["camera"] = e.camera_

        if e.getTags() and e.getTags():  # Assuming getTags is a method in the class
            data["tags"] = {key: value for key, value in e.getTags().items()}

        props = [[t.dev, t.prop, t.val] for t in e.properties_]
        if props:
            data["properties"] = props

        return json.dumps(data)

    @staticmethod
    def event_from_json(data, acq):
        if "special" in data:
            if data["special"] == "acquisition-end":
                return AcquisitionEvent.createAcquisitionFinishedEvent(acq)
            elif data["special"] == "sequence-end":
                return AcquisitionEvent.createAcquisitionSequenceEndEvent(acq)

        event = AcquisitionEvent(acq)

        if "axes" in data:
            for axisLabel, value in data["axes"].items():
                event.axisPositions_[axisLabel] = value

        if "min_start_time" in data:
            event.miniumumStartTime_ms_ = int(data["min_start_time"] * 1000)

        if "timeout" in data:
            event.timeout_ms_ = data["timeout"]

        if "config_group" in data:
            event.configGroup_ = data["config_group"][0]
            event.configPreset_ = data["config_group"][1]

        if "exposure" in data:
            event.exposure_ = data["exposure"]

        if "timeout_ms" in data:
            event.slmImage_ = data["timeout_ms"]

        if "stage_positions" in data:
            for stagePos in data["stage_positions"]:
                event.setStageCoordinate(stagePos[0], stagePos[1])

        if "z" in data:
            event.zPosition_ = data["z"]

        if "stage" in data:
            deviceName = data["stage"]["device_name"]
            position = data["stage"]["position"]
            event.axisPositions_[deviceName] = position
            if "axis_name" in data["stage"]:
                axisName = data["stage"]["axis_name"]
                event.stageDeviceNamesToAxisNames_[deviceName] = axisName

        # # Assuming XYTiledAcquisition is a class and AcqEngMetadata is a class or module with constants
        # if isinstance(event.acquisition_, XYTiledAcquisition):
        #     posIndex = event.acquisition_.getPixelStageTranslator().getPositionIndices(
        #         [int(event.axisPositions_[AcqEngMetadata.AXES_GRID_ROW])],
        #         [int(event.axisPositions_[AcqEngMetadata.AXES_GRID_COL])])[0]
        #     xyPos = event.acquisition_.getPixelStageTranslator().getXYPosition(posIndex).getCenter()
        #     event.xPosition_ = xyPos.x
        #     event.yPosition_ = xyPos.y

        if "x" in data:
            event.xPosition_ = data["x"]

        if "y" in data:
            event.yPosition_ = data["y"]

        if "slm_pattern" in data:
            event.slmImage_ = data["slm_pattern"]

        if "camera" in data:
            event.camera_ = data["camera"]

        if "tags" in data:
            tags = {key: value for key, value in data["tags"].items()}
            event.setTags(tags)

        if "properties" in data:
            for trip in data["properties"]:
                t = ThreeTuple(trip[0], trip[1], trip[2])
                event.properties_.add(t)

        return event

    def to_json(self):
        if self.sequence_:
            events = [self.event_to_json(e) for e in self.sequence_]
            return json.dumps({"events": events})
        else:
            return self.event_to_json(self)

    @staticmethod
    def from_json(data, acq):
        if "events" not in data:
            return AcquisitionEvent.event_from_json(data, acq)
        else:
            sequence = [AcquisitionEvent.event_from_json(item, acq) for item in data["events"]]
            return AcquisitionEvent(sequence)

    def get_camera_device_name(self):
        return self.camera_

    def set_camera_device_name(self, camera):
        self.camera_ = camera

    def get_additional_properties(self):
        return [(t.dev, t.prop, t.val) for t in self.properties_]

    def should_acquire_image(self):
        if self.sequence_:
            return True
        return self.configPreset_ is not None or len(self.axisPositions_) > 0

    def has_config_group(self):
        return self.configPreset_ is not None and self.configGroup_ is not None

    def get_config_preset(self):
        return self.configPreset_

    def get_config_group(self):
        return self.configGroup_

    def set_config_preset(self, config):
        self.configPreset_ = config

    def set_config_group(self, group):
        self.configGroup_ = group

    def get_exposure(self):
        return self.exposure_

    def set_exposure(self, exposure):
        self.exposure_ = exposure

    def set_property(self, device, property, value):
        self.properties_.add(ThreeTuple(device, property, value))

    def set_minimum_start_time(self, l):
        self.miniumumStartTime_ms_ = l

    def get_defined_axes(self):
        return set(self.axisPositions_.keys())

    def set_axis_position(self, label, position):
        if position is None:
            raise Exception("Cannot set axis position to null")
        self.axisPositions_[label] = position

    def set_stage_coordinate(self, deviceName, v, axisName=None):
        self.stageCoordinates_[deviceName] = v
        self.stageDeviceNamesToAxisNames_[deviceName] = deviceName if axisName is None else axisName

    def get_stage_single_axis_stage_position(self, deviceName):
        return self.stageCoordinates_.get(deviceName)

    def get_axis_positions(self):
        return self.axisPositions_

    def get_axis_position(self, label):
        return self.axisPositions_.get(label)

    def get_timeout_ms(self):
        return self.timeout_ms_

    def set_time_index(self, index):
        self.set_axis_position(AcqEngMetadata.TIME_AXIS, index)

    def set_channel_name(self, name):
        self.set_axis_position(AcqEngMetadata.CHANNEL_AXIS, name)

    def get_slm_image(self):
        return self.slmImage_

    def set_z(self, index, position):
        if index is not None:
            self.set_axis_position(AcqEngMetadata.Z_AXIS, index)
        self.zPosition_ = position

    def get_t_index(self):
        return self.get_axis_position(AcqEngMetadata.TIME_AXIS)

    def get_z_index(self):
        return self.get_axis_position(AcqEngMetadata.Z_AXIS)

    def get_device_axis_name(self, deviceName):
        if deviceName not in self.stageDeviceNamesToAxisNames_:
            raise Exception(f"No axis name for device {deviceName}. call setStageCoordinate first")
        return self.stageDeviceNamesToAxisNames_[deviceName]

    def get_stage_device_names(self):
        return set(self.stageDeviceNamesToAxisNames_.keys())

    @staticmethod
    def create_acquisition_finished_event(acq):
        evt = AcquisitionEvent(acq)
        evt.specialFlag_ = AcquisitionEvent.SpecialFlag.ACQUISITION_FINISHED
        return evt

    def is_acquisition_finished_event(self):
        return self.specialFlag_ == AcquisitionEvent.SpecialFlag.ACQUISITION_FINISHED

    @staticmethod
    def create_acquisition_sequence_end_event(acq):
        evt = AcquisitionEvent(acq)
        evt.specialFlag_ = AcquisitionEvent.SpecialFlag.ACQUISITION_SEQUENCE_END
        return evt

    def is_acquisition_sequence_end_event(self):
        return self.specialFlag_ == AcquisitionEvent.SpecialFlag.ACQUISITION_SEQUENCE_END

    def get_z_position(self):
        return self.zPosition_

    def get_minimum_start_time_absolute(self):
        if self.miniumumStartTime_ms_ is None:
            return None
        return self.acquisition_.get_start_time_ms() + self.miniumumStartTime_ms_

    def get_sequence(self):
        return self.sequence_

    def is_exposure_sequenced(self):
        return self.exposureSequenced_

    def is_config_group_sequenced(self):
        return self.configGroupSequenced_

    def is_xy_sequenced(self):
        return self.xySequenced_

    def is_z_sequenced(self):
        return self.zSequenced_

    def get_x_position(self):
        return self.xPosition_

    def get_camera_image_counts(self, default_camera_device_name):
        """
        Get the number of images to be acquired on each camera in a sequence event.
        For a non-sequence event, the number of images is 1, and the camera is the core camera.
        This is passed in as an argument in order to avoid this class talking to the core directly.

        Args:
            default_camera_device_name (str): Default camera device name.

        Returns:
            defaultdict: Dictionary containing the camera device names as keys and image counts as values.
        """
        # Figure out how many images on each camera and start sequence with appropriate number on each
        camera_image_counts = {}
        camera_device_names = set()
        if self.get_sequence() is None:
            camera_image_counts[default_camera_device_name] = 1
            return camera_image_counts

        for event in self.get_sequence():
            camera_device_names.add(event.get_camera_device_name() if event.get_camera_device_name() is not None else
                                    default_camera_device_name)
        if None in camera_device_names:
            camera_device_names.remove(None)
            camera_device_names.add(default_camera_device_name)

        for camera_device_name in camera_device_names:
            camera_image_counts[camera_device_name] = sum(1 for event in self.get_sequence()
                                                          if event.get_camera_device_name() == camera_device_name)

            if len(camera_device_names) == 1 and camera_device_name == default_camera_device_name:
                camera_image_counts[camera_device_name] = len(self.get_sequence())

        return camera_image_counts

    def get_y_position(self):
        return self.yPosition_

    def get_position_name(self):
        axisPosition_ = self.get_axis_position(AcqEngMetadata.POSITION_AXIS)
        if isinstance(axisPosition_, str):
            return axisPosition_
        return None

    def set_x(self, x):
        self.xPosition_ = x

    def set_y(self, y):
        self.yPosition_ = y

    def set_tags(self, tags):
        self.tags_.clear()
        if tags:
            self.tags_.update(tags)

    def get_tags(self):
        return dict(self.tags_)

    def __str__(self):
        if self.specialFlag_ == AcquisitionEvent.SpecialFlag.AcquisitionFinished:
            return "Acq finished event"
        elif self.specialFlag_ == AcquisitionEvent.SpecialFlag.AcquisitionSequenceEnd:
            return "Acq sequence end event"

        builder = []
        for deviceName in self.stageDeviceNamesToAxisNames_.keys():
            builder.append(f"\t{deviceName}: {self.get_stage_single_axis_stage_position(deviceName)}")

        if self.zPosition_ is not None:
            builder.append(f"z {self.zPosition_}")
        if self.xPosition_ is not None:
            builder.append(f"x {self.xPosition_}")
        if self.yPosition_ is not None:
            builder.append(f"y {self.yPosition_}")

        for axis in self.axisPositions_.keys():
            builder.append(f"\t{axis}: {self.axisPositions_[axis]}")

        if self.camera_ is not None:
            builder.append(f"\t{self.camera_}: {self.camera_}")

        return ' '.join(builder)


ThreeTuple = namedtuple('ThreeTuple', ['dev', 'prop', 'val'])



import json

class AcqNotification:

    class Acquisition:
        ACQ_STARTED = "acq_started"
        ACQ_EVENTS_FINISHED = "acq_events_finished"

        @staticmethod
        def to_string():
            return "global"

    class Hardware:
        PRE_HARDWARE = "pre_hardware"
        POST_HARDWARE = "post_hardware"

        @staticmethod
        def to_string():
            return "hardware"

    class Camera:
        PRE_SEQUENCE_STARTED = "pre_sequence_started"
        PRE_SNAP = "pre_snap"
        POST_EXPOSURE = "post_exposure"

        @staticmethod
        def to_string():
            return "camera"

    class Image:
        IMAGE_SAVED = "image_saved"
        DATA_SINK_FINISHED = "data_sink_finished"

        @staticmethod
        def to_string():
            return "image"

    def __init__(self, type, id, phase=None):
        if type == AcqNotification.Acquisition or type == AcqNotification.Acquisition.to_string():
            self.type = AcqNotification.Acquisition
            self.id = id
            self.phase = phase
        elif (type == AcqNotification.Image or type == AcqNotification.Image.to_string()) and \
              phase == AcqNotification.Image.DATA_SINK_FINISHED:
            self.type = AcqNotification.Image
            self.id = id
            self.phase = phase
        elif phase in [AcqNotification.Camera.PRE_SNAP, AcqNotification.Camera.POST_EXPOSURE,
                     AcqNotification.Camera.PRE_SEQUENCE_STARTED]:
            self.type = AcqNotification.Camera
            self.id = json.loads(id) if isinstance(id, str) else id # convert from '{'time': 5}' to {'time': 5}
        elif phase in [AcqNotification.Hardware.PRE_HARDWARE, AcqNotification.Hardware.POST_HARDWARE]:
            self.type = AcqNotification.Hardware
            self.id = json.loads(id) if isinstance(id, str) else id # convert from '{'time': 5}' to {'time': 5}
        elif phase == AcqNotification.Image.IMAGE_SAVED:
            self.type = AcqNotification.Image
            self.id = id
        else:
            raise ValueError("Unknown phase")
        self.phase = phase


    @staticmethod
    def create_acq_events_finished_notification():
       return AcqNotification(AcqNotification.Acquisition, None, AcqNotification.Acquisition.ACQ_EVENTS_FINISHED)

    @staticmethod
    def create_acq_started_notification():
        return AcqNotification(AcqNotification.Acquisition, None, AcqNotification.Acquisition.ACQ_STARTED)

    @staticmethod
    def create_data_sink_finished_notification():
        return AcqNotification(AcqNotification.Image, None, AcqNotification.Image.DATA_SINK_FINISHED)

    @staticmethod
    def create_image_saved_notification(image_descriptor):
        return AcqNotification(AcqNotification.Image, image_descriptor, AcqNotification.Image.IMAGE_SAVED)

    def to_json(self):
        n = {}
        n['type'] = self.type
        n['phase'] = self.phase
        if self.id:
            n['id'] = self.id
        return n

    @staticmethod
    def from_json(json):
        return AcqNotification(json['type'],
                               json['id'] if 'id' in json else None,
                               json['phase'] if 'phase' in json else None)

    def is_acquisition_finished_notification(self):
        return self.phase == AcqNotification.Acquisition.ACQ_EVENTS_FINISHED

    def is_data_sink_finished_notification(self):
        return self.phase == AcqNotification.Image.DATA_SINK_FINISHED

    def is_image_saved_notification(self):
        return self.phase == AcqNotification.Image.IMAGE_SAVED



"""
The Pycro-manager Acquisiton system
"""
import warnings
import weakref

import numpy as np
import multiprocessing
import threading
from inspect import signature
import time
from pyjavaz import deserialize_array
from pyjavaz import PullSocket, PushSocket, JavaObject, JavaClass
from pyjavaz import DEFAULT_BRIDGE_PORT as DEFAULT_PORT
from pycromanager.mm_java_classes import ZMQRemoteMMCoreJ, Magellan
import pycromanager.logging as logging
from ndtiff import Dataset
import os.path
import queue
from docstring_inheritance import NumpyDocstringInheritanceMeta
from pycromanager.acquisition.acquisition_superclass import Acquisition
import traceback
from pycromanager.acq_future import AcqNotification, AcquisitionFuture



### These functions are defined outside the Acquisition class to
# prevent problems with pickling when running them in differnet process
# although they are currently only used in different threads

def _run_acq_event_source(acquisition, event_port, event_queue, debug=False):  
    event_socket = PushSocket(event_port, debug=debug)
    try:
        while True:
            events = event_queue.get(block=True)
            if debug:
                logging.main_logger.debug(f"got event(s): {events}")
            if events is None:
                # Initiate the normal shutdown process
                if not acquisition._acq.is_finished():
                    # if it has been finished through something happening on the other side
                    event_socket.send({"events": [{"special": "acquisition-end"}]})
                    # wait for signal that acquisition has received the end signal
                    while not acquisition._acq.is_finished():
                        acquisition._acq.block_until_events_finished(0.01)
                break
            # it may have been shut down remotely (e.g. by user Xing out viewer)
            # if we try to send an event at this time, it will hang indefinitely
            if acquisition._acq.is_finished():
                break
            # TODO in theory it could be aborted in between the check above and sending below,
            #  maybe consider putting a timeout on the send?
            event_socket.send({"events": events if type(events) == list else [events]})
            if debug:
                logging.main_logger.debug("sent events")
    except Exception as e:
        acquisition.abort(e)
    finally:
        event_socket.close()


def _run_acq_hook(acquisition, pull_port,
                  push_port, hook_connected_evt, event_queue, hook_fn, debug=False):

    push_socket = PushSocket(pull_port, debug=debug)
    pull_socket = PullSocket(push_port, debug=debug)
    hook_connected_evt.set()

    exception = None
    while True:
        event_msg = pull_socket.receive()

        if "special" in event_msg and event_msg["special"] == "acquisition-end":
            push_socket.send(event_msg)
            push_socket.close()
            pull_socket.close()
            return
        else:
            if "events" in event_msg.keys():
                event_msg = event_msg["events"]  # convert from sequence
            params = signature(hook_fn).parameters
            if len(params) == 1 or len(params) == 2:
                try:
                    if len(params) == 1:
                        new_event_msg = hook_fn(event_msg)
                    elif len(params) == 2:
                        new_event_msg = hook_fn(event_msg, event_queue)
                except Exception as e:
                    acquisition.abort(e)
                    # Cancel the execution of event because there was an exception
                    new_event_msg = None
                    # don't return here--wait for the signal from the acq engine
            else:
                acquisition.abort(Exception("Incorrect number of arguments for hook function. Must be 1 or 2"))

        if isinstance(new_event_msg, list):
            new_event_msg = {
                "events": new_event_msg
            }  # convert back to the expected format for a sequence
        push_socket.send(new_event_msg)


def _run_image_processor(
        acquisition, pull_port, push_port, sockets_connected_evt, process_fn, event_queue, debug
):
    acquisition._process_fn = process_fn
    push_socket = PushSocket(pull_port, debug=debug)
    pull_socket = PullSocket(push_port, debug=debug)
    if debug:
        logging.main_logger.debug("image processing sockets connected")
    sockets_connected_evt.set()

    def process_and_sendoff(image_tags_tuple, original_dtype):
        """

        Parameters
        ----------
        image_tags_tuple :


        Returns
        -------

        """
        if len(image_tags_tuple) != 2:
            acquisition.abort(Exception("If image is returned, it must be of the form (pixel, metadata)"))

        pixels = image_tags_tuple[0]
        metadata = image_tags_tuple[1]

        # only accepts same pixel type as original
        if not np.issubdtype(image_tags_tuple[0].dtype, original_dtype) and not np.issubdtype(
            original_dtype, image_tags_tuple[0].dtype
        ):
            acquisition.abort(Exception(
                "Processed image pixels must have same dtype as input image pixels, "
                "but instead they were {} and {}".format(image_tags_tuple[0].dtype, pixels.dtype)
            ))

        if metadata['PixelType'] == 'RGB32':
            if pixels.shape[-1] == 3:
                #append 0 for alpha channel because thats whats expected
                pixels = np.concatenate([pixels, np.zeros_like(pixels[..., 0])[..., None]], axis=2)
        else:
            #maybe pixel type was changed by processing?
            metadata["PixelType"] = "GRAY8" if pixels.dtype.itemsize == 1 else "GRAY16"

        processed_img = {
            "pixels": pixels.tobytes(),
            "metadata": metadata,
        }
        push_socket.send(processed_img, suppress_debug_message=True)

    while True:
        message = None
        while message is None:
            message = pull_socket.receive(timeout=30, suppress_debug_message=True)  # check for new message

        if "special" in message and message["special"] == "finished":
            pull_socket.close()
            push_socket.send(message)  # Continue propagating the finished signal
            push_socket.close()
            return

        metadata = message["metadata"]
        # TODO: this should probably be handled by the push socket...
        pixels = deserialize_array(message["pixels"])
        if metadata['PixelType'] == 'RGB32':
            image = np.reshape(pixels, [metadata["Height"], metadata["Width"], 4])[..., :3]
        else:
            image = np.reshape(pixels, [metadata["Height"], metadata["Width"]])

        processed = acquisition._call_image_process_fn(image, metadata)

        if processed is None:
            continue

        if type(processed) == list:
            for image in processed:
                process_and_sendoff(image, pixels.dtype)
        else:
            process_and_sendoff(processed, pixels.dtype)

def _notification_handler_fn(acquisition, notification_push_port, connected_event, debug=False):
    monitor_socket = PullSocket(notification_push_port)
    connected_event.set()

    try:
        events_finished = False
        data_sink_finished = False
        while True:
            message = monitor_socket.receive()
            notification = AcqNotification.from_json(message)
            # these are processed seperately to handle image saved callback
            if AcqNotification.is_image_saved_notification(notification):
                if not notification.is_data_sink_finished_notification():
                    # decode the NDTiff index entry
                    index_entry = notification.id.encode('ISO-8859-1')
                    axes = acquisition._dataset._add_index_entry(index_entry)
                    # swap the notification id from the byte array of index information to axes
                    notification.id = axes
                acquisition._image_notification_queue.put(notification)
            acquisition._notification_queue.put(notification)
            if AcqNotification.is_acquisition_finished_notification(notification):
                events_finished = True
            elif AcqNotification.is_data_sink_finished_notification(notification):
                data_sink_finished = True
                acquisition._image_notification_queue.put(notification)
            if events_finished and data_sink_finished:
                break

    except Exception as e:
        traceback.print_exc()
        acquisition.abort(e)
    finally:
        monitor_socket.close()


class JavaBackendAcquisition(Acquisition, metaclass=NumpyDocstringInheritanceMeta):
    """
    Pycro-Manager acquisition that uses a Java runtime backend via a ZeroMQ communication layer.
    """

    def __init__(
        self,
        directory: str=None,
        name: str='default_acq_name',
        image_process_fn : callable=None,
        event_generation_hook_fn: callable=None,
        pre_hardware_hook_fn: callable=None,
        post_hardware_hook_fn: callable=None,
        post_camera_hook_fn: callable=None,
        notification_callback_fn: callable=None,
        image_saved_fn: callable=None,
        show_display: bool=True,
        napari_viewer=None,
        saving_queue_size: int=20,
        timeout: int=2500,
        port: int=DEFAULT_PORT,
        debug: int=False
    ):
        # Get a dict of all named argument values (or default values when nothing provided)
        arg_names = [k for k in signature(JavaBackendAcquisition.__init__).parameters.keys() if k != 'self']
        l = locals()
        named_args = {arg_name: (l[arg_name] if arg_name in l else
                                     dict(signature(JavaBackendAcquisition.__init__).parameters.items())[arg_name].default)
                                     for arg_name in arg_names }


        superclass_arg_names = [k for k in signature(Acquisition.__init__).parameters.keys() if k != 'self']
        superclass_args = {key: named_args[key] for key in superclass_arg_names}
        super().__init__(**superclass_args)

        if directory is not None:
            # Expend ~ in path
            directory = os.path.expanduser(directory)
            # If path is relative, retain knowledge of the current working directory
            self._directory = os.path.abspath(directory)
        else:
            self._directory = None
        named_args['directory'] = self._directory

        # Java specific parameters
        self._port = port
        self._timeout = timeout
        self._nd_viewer = None

        self._create_event_queue()
        self._create_remote_acquisition(**named_args)
        self._initialize_image_processor(**named_args)
        self._initialize_hooks(**named_args)

        try:
            self._remote_notification_handler = JavaObject('org.micromanager.remote.RemoteNotificationHandler',
                                                           args=[self._acq], port=self._port, new_socket=False,
                                                           timeout=self._timeout)
            self._acq_notification_recieving_thread = self._start_receiving_notifications()
            self._acq_notification_dispatcher_thread = self._start_notification_dispatcher(notification_callback_fn)
        # TODO: can remove this after this feature has been present for a while
        except:
            traceback.print_exc()
            warnings.warn('Could not create acquisition notification handler. '
                          'Update Micro-Manager and Pyrcro-Manager to the latest versions to fix this')

        # Start remote acquisition
        # Acquistition.start is now deprecated, so this can be removed later
        # Acquisitions now get started automatically when the first events submitted
        # but Magellan acquisitons (and probably others that generate their own events)
        # will need some new method to submit events only after image processors etc have been added
        self._acq.start()
        self._dataset_disk_location = (
            self._acq.get_data_sink().get_storage().get_disk_location()
            if self._acq.get_data_sink() is not None
            else None
        )

        self._start_events()

        # Load remote storage
        data_sink = self._acq.get_data_sink()
        if data_sink is not None:
            # load a view of the dataset in progress. This is used so that acq.get_dataset() can be called
            # while the acquisition is still running, and (optionally )so that a image_saved_fn can be called
            # when images are written to disk
            ndtiff_storage = data_sink.get_storage()
            summary_metadata = ndtiff_storage.get_summary_metadata()
            if directory is not None:
                self._dataset = Dataset(dataset_path=self._dataset_disk_location, _summary_metadata=summary_metadata)
                # Monitor image arrival so they can be loaded on python side, but with no callback function
                # Need to do this regardless of whether you use it, so that it signals to shut down on Java side
                self._storage_monitor_thread = self._add_storage_monitor_fn(image_saved_fn=image_saved_fn)

        if show_display:
            if napari_viewer is None:
                # using NDViewer
                self._nd_viewer = self._acq.get_data_sink().get_viewer()
            else:
                # using napari viewer
                try:
                    import napari
                except:
                    raise Exception('Napari must be installed in order to use this feature')
                from pycromanager.napari_util import start_napari_signalling
                assert isinstance(napari_viewer, napari.Viewer), 'napari_viewer must be an instance of napari.Viewer'
                self._napari_viewer = napari_viewer
                start_napari_signalling(self._napari_viewer, self.get_dataset())


    ########  Public API methods with unique implementations for Java backend ###########
    def get_dataset(self):
        if self._finished:
            if self._dataset is None:
                self._dataset = Dataset(self._dataset_disk_location)

        return self._dataset

    def await_completion(self):
        while not self._acq.are_events_finished() or (
                self._acq.get_data_sink() is not None and not self._acq.get_data_sink().is_finished()):
            self._check_for_exceptions()
            self._acq.block_until_events_finished(0.01)
        # This will block until saving is finished, if there is a data sink
        self._acq.wait_for_completion()
        self._check_for_exceptions()

        for hook_thread in self._hook_threads:
            hook_thread.join()

        if hasattr(self, '_event_thread'):
            self._event_thread.join()

        # need to do this so its _Bridge can be garbage collected and a reference to the JavaBackendAcquisition
        # does not prevent Bridge cleanup and process exiting
        self._remote_acq = None

        # Wait on all the other threads to shut down properly
        if hasattr(self, '_storage_monitor_thread'):
            self._storage_monitor_thread.join()

        if hasattr(self, '_acq_notification_recieving_thread'):
            # for backwards compatiblitiy with older versions of Pycromanager java before this added
            self._acq_notification_recieving_thread.join()
            self._remote_notification_handler.notification_handling_complete()
            # need to do this so its _Bridge can be garbage collected and a reference to the JavaBackendAcquisition
            # does not prevent Bridge cleanup and process exiting
            self._remote_notification_handler = None
            self._acq_notification_dispatcher_thread.join()

        self._acq = None
        self._finished = True


    def get_viewer(self):
        if self._napari_viewer is None:
            return self._nd_viewer
        else:
            return self._napari_viewer

    ########  Private methods ###########
    def _start_receiving_notifications(self):
        """
        Thread that runs a function that pulls notifications from the acquisition engine and puts them on a queue
        """
        connected_event = threading.Event()

        pull_port = self._remote_notification_handler.get_port()
        notification_thread = threading.Thread(
            target=_notification_handler_fn,
            args=(
                self,
                pull_port,
                connected_event,
                self._debug,
            ),
            name="NotificationHandlerThread",
        )
        #
        # Wait for pulling to start before you signal for pushing to start
        notification_thread.start()
        connected_event.wait()

        # start pushing out all the notifications
        self._remote_notification_handler.start()
        return notification_thread

    def _add_storage_monitor_fn(self, image_saved_fn=None):
        """
        Add a callback function that gets called whenever a new image is writtern to disk (for acquisitions in
        progress only)

        Parameters
        ----------
        image_saved_fn : Callable
            user function to be run whenever an image is ready on disk
        """
        # TODO: this should read from a queue of image-specific notifications and dispatch accordingly

        callback = None
        if image_saved_fn is not None:
            params = signature(image_saved_fn).parameters
            if len(params) == 2:
                callback = image_saved_fn
            elif len(params) == 3:
                callback = lambda axes, dataset: image_saved_fn(axes, dataset, self._event_queue)
            else:
                raise Exception('Image saved callbacks must have either 2 or three parameters')

        def _storage_monitor_fn():
            dataset = self.get_dataset()
            while True:
                image_notification = self._image_notification_queue.get()
                if AcqNotification.is_data_sink_finished_notification(image_notification):
                    break
                dataset._new_image_arrived = True
                if callback is not None:
                    callback(image_notification.id, dataset)
        t = threading.Thread(target=_storage_monitor_fn, name='StorageMonitorThread')
        t.start()
        return t

    def _check_for_exceptions(self):
        """
        Check for exceptions on the python side (i.e. hooks and processors)
        or on the Java side (i.e. hardware control)
        """
        # these will throw exceptions
        self._acq.check_for_exceptions()
        if self._exception is not None:
            raise self._exception

    def _start_events(self, **kwargs):

        self.event_port = self._acq.get_event_port()

        self._event_thread = threading.Thread(
            target=_run_acq_event_source,
            args=(self, self.event_port, self._event_queue, self._debug),
            name="Event sending",
        )
        self._event_thread.start()

    def _initialize_image_processor(self, **kwargs):

        if kwargs['image_process_fn'] is not None:
            java_processor = JavaObject(
                "org.micromanager.remote.RemoteImageProcessor", port=self._port, timeout=self._timeout
            )
            self._acq.add_image_processor(java_processor)
            self._processor_thread = self._start_processor(
                java_processor, kwargs['image_process_fn'],
                # Some acquisitions (e.g. Explore acquisitions) create events on Java side
                self._event_queue if hasattr(self, '_event_queue') else None,
                process=False)

    def _initialize_hooks(self, **kwargs):
        self._hook_threads = []
        if kwargs['event_generation_hook_fn'] is not None:
            hook = JavaObject(
                "org.micromanager.remote.RemoteAcqHook", port=self._port, args=[self._acq], timeout=self._timeout
            )
            self._hook_threads.append(self._start_hook(hook, kwargs['event_generation_hook_fn'],
                                                       self._event_queue, process=False))
            self._acq.add_hook(hook, self._acq.EVENT_GENERATION_HOOK)
        if kwargs['pre_hardware_hook_fn'] is not None:
            hook = JavaObject(
                "org.micromanager.remote.RemoteAcqHook", port=self._port, args=[self._acq], timeout=self._timeout
            )
            self._hook_threads.append(self._start_hook(hook,
                                            kwargs['pre_hardware_hook_fn'], self._event_queue,
                                                       process=False))
            self._acq.add_hook(hook, self._acq.BEFORE_HARDWARE_HOOK)
        if kwargs['post_hardware_hook_fn'] is not None:
            hook = JavaObject(
                "org.micromanager.remote.RemoteAcqHook", port=self._port, args=[self._acq], timeout=self._timeout
            )
            self._hook_threads.append(self._start_hook(hook, kwargs['post_hardware_hook_fn'],
                                                       self._event_queue, process=False))
            self._acq.add_hook(hook, self._acq.AFTER_HARDWARE_HOOK)
        if kwargs['post_camera_hook_fn'] is not None:
            hook = JavaObject(
                "org.micromanager.remote.RemoteAcqHook", port=self._port, args=[self._acq], timeout=self._timeout
            )
            self._hook_threads.append(self._start_hook(hook, kwargs['post_camera_hook_fn'],
                                                       self._event_queue, process=False))
            self._acq.add_hook(hook, self._acq.AFTER_CAMERA_HOOK)

    def _create_remote_acquisition(self, **kwargs):
        core = ZMQRemoteMMCoreJ(port=self._port, timeout=self._timeout, debug=self._debug)
        acq_factory = JavaObject("org.micromanager.remote.RemoteAcquisitionFactory",
            # create a new socket for it to run on so that it can have blocking calls without interfering with
            # the main socket or other internal sockets
            new_socket=True,
            port=self._port, args=[core], debug=self._debug, timeout=self._timeout)
        show_viewer = kwargs['show_display'] is True and kwargs['napari_viewer'] is None
        self._acq = acq_factory.create_acquisition(kwargs['directory'], kwargs['name'], show_viewer,
                                                   kwargs['saving_queue_size'], self._debug,)

    def _start_hook(self, remote_hook, remote_hook_fn : callable, event_queue, process):
        """

        Parameters
        ----------
        remote_hook :

        remote_hook_fn :

        event_queue :

        process :


        Returns
        -------

        """
        hook_connected_evt = multiprocessing.Event() if process else threading.Event()

        pull_port = remote_hook.get_pull_port()
        push_port = remote_hook.get_push_port()

        hook_thread = (multiprocessing.Process if process else threading.Thread)(
            target=_run_acq_hook,
            name="AcquisitionHook",
            args=(
                self,
                pull_port,
                push_port,
                hook_connected_evt,
                event_queue,
                remote_hook_fn,
                self._debug,
            ),
        )
        # if process else threading.Thread(target=_acq_hook_fn, args=(), name='AcquisitionHook')
        hook_thread.start()

        hook_connected_evt.wait()  # wait for push/pull sockets to connect
        return hook_thread

    def _start_processor(self, processor, process_fn, event_queue, process):
        """

        Parameters
        ----------
        processor :

        process_fn :

        event_queue :

        process :


        Returns
        -------

        """
        # this must start first
        processor.start_pull()

        sockets_connected_evt = multiprocessing.Event() if process else threading.Event()

        pull_port = processor.get_pull_port()
        push_port = processor.get_push_port()

        processor_thread = (multiprocessing.Process if process else threading.Thread)(
            target=_run_image_processor,
            args=(
                self,
                pull_port,
                push_port,
                sockets_connected_evt,
                process_fn,
                event_queue,
                self._debug,
            ),
            name="ImageProcessor",
        )
        processor_thread.start()

        sockets_connected_evt.wait()  # wait for push/pull sockets to connect
        processor.start_push()
        return processor_thread


class XYTiledAcquisition(JavaBackendAcquisition):
    """
    For making tiled images with an XY stage and multiresolution saving
    (e.g. for making one large contiguous image of a sample larger than the field of view)
    """

    def __init__(
            self,
            tile_overlap : int or tuple,
            directory: str=None,
            name: str=None,
            max_multi_res_index: int=None,
            image_process_fn: callable=None,
            pre_hardware_hook_fn: callable=None,
            post_hardware_hook_fn: callable=None,
            post_camera_hook_fn: callable=None,
            show_display: bool=True,
            image_saved_fn: callable=None,
            saving_queue_size: int=20,
            timeout: int=1000,
            port: int=DEFAULT_PORT,
            debug: bool=False,
    ):
        """
        Parameters
        ----------
        tile_overlap : int or tuple of int
            If given, XY tiles will be laid out in a grid and multi-resolution saving will be
            actived. Argument can be a two element tuple describing the pixel overlaps between adjacent
            tiles. i.e. (pixel_overlap_x, pixel_overlap_y), or an integer to use the same overlap for both.
            For these features to work, the current hardware configuration must have a valid affine transform
            between camera coordinates and XY stage coordinates
        max_multi_res_index : int
            Maximum index to downsample to in multi-res pyramid mode. 0 is no downsampling,
            1 is downsampled up to 2x, 2 is downsampled up to 4x, etc. If not provided, it will be dynamically
            calculated and updated from data
        """
        self.tile_overlap = tile_overlap
        self.max_multi_res_index = max_multi_res_index
        # Collct all argument values except the ones specific to XY Tiled acquisitions
        arg_names = list(signature(self.__init__).parameters.keys())
        arg_names.remove('tile_overlap')
        arg_names.remove('max_multi_res_index')
        l = locals()
        named_args = {arg_name: l[arg_name] for arg_name in arg_names}
        super().__init__(**named_args)

    def _create_remote_acquisition(self, port, **kwargs):
        core = ZMQRemoteMMCoreJ(port=self._port, timeout=self._timeout)
        acq_factory = JavaObject("org.micromanager.remote.RemoteAcquisitionFactory",
                                 port=self._port, args=[core], timeout=self._timeout)

        show_viewer = kwargs['show_display'] is True and\
                      kwargs['napari_viewer'] is None and\
                      (kwargs['directory'] is not None and kwargs['name'] is not None)
        if type(self.tile_overlap) is tuple:
            x_overlap, y_overlap = self.tile_overlap
        else:
            x_overlap = self.tile_overlap
            y_overlap = self.tile_overlap

        self._remote_acq = acq_factory.create_tiled_acquisition(
            kwargs['directory'],
            kwargs['name'],
            show_viewer,
            True,
            x_overlap,
            y_overlap,
            self.max_multi_res_index if self.max_multi_res_index is not None else -1,
            kwargs['saving_queue_size'],
            self._debug,
        )

class ExploreAcquisition(JavaBackendAcquisition):
    """
    Launches a user interface for an "Explore Acquisition"--a type of XYTiledAcquisition
    in which acquisition events come from the user dynamically driving the stage and selecting
    areas to image
    """

    def __init__(
            self,
            directory: str,
            name: str,
            z_step_um: float,
            tile_overlap: int or tuple,
            channel_group: str = None,
            image_process_fn: callable=None,
            pre_hardware_hook_fn: callable=None,
            post_hardware_hook_fn: callable=None,
            post_camera_hook_fn: callable=None,
            show_display: bool=True,
            image_saved_fn: callable=None,
            saving_queue_size: int=20,
            timeout: int=2500,
            port: int=DEFAULT_PORT,
            debug: bool=False,
    ):
        """
        Parameters
        ----------
        z_step_um : str
            Spacing between successive z planes, in microns
        tile_overlap : int or tuple of int
            If given, XY tiles will be laid out in a grid and multi-resolution saving will be
            activated. Argument can be a two element tuple describing the pixel overlaps between adjacent
            tiles. i.e. (pixel_overlap_x, pixel_overlap_y), or an integer to use the same overlap for both.
            For these features to work, the current hardware configuration must have a valid affine transform
            between camera coordinates and XY stage coordinates
        channel_group : str
            Name of a config group that provides selectable channels
        """
        self.tile_overlap = tile_overlap
        self.channel_group = channel_group
        self.z_step_um = z_step_um
        # Collct all argument values except the ones specific to ExploreAcquisitions
        arg_names = list(signature(self.__init__).parameters.keys())
        arg_names.remove('tile_overlap')
        arg_names.remove('z_step_um')
        arg_names.remove('channel_group')
        l = locals()
        named_args = {arg_name: l[arg_name] for arg_name in arg_names}
        super().__init__(**named_args)

    def _create_remote_acquisition(self, port, **kwargs):
        if type(self.tile_overlap) is tuple:
            x_overlap, y_overlap = self.tile_overlap
        else:
            x_overlap = self.tile_overlap
            y_overlap = self.tile_overlap

        ui_class = JavaClass('org.micromanager.explore.ExploreAcqUIAndStorage')
        ui = ui_class.create(kwargs['directory'], kwargs['name'], x_overlap, y_overlap, self.z_step_um, self.channel_group)
        self._remote_acq = ui.get_acquisition()

    def _start_events(self, **kwargs):
        pass # These come from the user

    def _create_event_queue(self, **kwargs):
        pass # Comes from the user


class MagellanAcquisition(JavaBackendAcquisition):
    """
    Class used for launching Micro-Magellan acquisitions. Must pass either magellan_acq_index
    or magellan_explore as an argument
    """

    def __init__(
            self,
            magellan_acq_index: int=None,
            magellan_explore: bool=False,
            image_process_fn: callable=None,
            event_generation_hook_fn: callable=None,
            pre_hardware_hook_fn: callable=None,
            post_hardware_hook_fn: callable=None,
            post_camera_hook_fn: callable=None,
            image_saved_fn: callable=None,
            timeout: int=500,
            port: int=DEFAULT_PORT,
            debug: bool=False,
    ):
        """
        Parameters
        ----------
        magellan_acq_index : int
            run this acquisition using the settings specified at this position in the main
            GUI of micro-magellan (micro-manager plugin). This index starts at 0
        magellan_explore : bool
            Run a Micro-magellan explore acquisition
        """
        self.magellan_acq_index = magellan_acq_index
        self.magellan_explore = magellan_explore
        # Collct all argument values except the ones specific to Magellan
        arg_names = list(signature(self.__init__).parameters.keys())
        arg_names.remove('magellan_acq_index')
        arg_names.remove('magellan_explore')
        l = locals()
        named_args = {arg_name: l[arg_name] for arg_name in arg_names}
        super().__init__(**named_args)

    def _start_events(self, **kwargs):
        pass # Magellan handles this on Java side

    def _create_event_queue(self, **kwargs):
        pass # Magellan handles this on Java side

    def _create_remote_acquisition(self, **kwargs):
        magellan_api = Magellan()
        if self.magellan_acq_index is not None:
            self._remote_acq = magellan_api.create_acquisition(self.magellan_acq_index, False)
        elif self.magellan_explore:
            self._remote_acq = magellan_api.create_explore_acquisition(False)
        self._event_queue = None



import subprocess
import platform
import atexit
import threading
import types

from pycromanager.acquisition.acq_eng_py.internal.engine import Engine
from pymmcore import CMMCore
import pycromanager.logging as logging
import pymmcore
from pyjavaz import DEFAULT_BRIDGE_PORT, server_terminated

import re

class TaggedImage:

    def __init__(self, tags, pix):
        self.tags = tags
        self.pix = pix

def _camel_to_snake(name):
    """
    Convert camelCase string to snake_case
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

def _create_pymmcore_instance():
    """
    Make a subclass of CMMCore with two differences:

    1. All methods are converted to snake_case
    2. add convenience methods to match the MMCoreJ API:
    """

    # Create a new dictionary for the class attributes
    new_attributes = {}

    # Iterate through the original attributes
    for attr_name, attr_value in vars(CMMCore).items():
        # If it's a dunder method, skip it (we don't want to override these)
        if attr_name.startswith("__") and attr_name.endswith("__"):
            continue
        # If the attribute is callable (i.e., a method), convert its name to snake_case and add it
        if callable(attr_value):
            new_attr_name = _camel_to_snake(attr_name)
            new_attributes[new_attr_name] = attr_value

    # Create and return a new class that subclasses the original class and has the new attributes
    clz = type(CMMCore.__name__ + "SnakeCase", (CMMCore,), new_attributes)

    instance = clz()

    def pop_next_tagged_image(self):
        md = pymmcore.Metadata()
        pix = self.pop_next_image_md(0, 0, md)
        tags = {key: md.GetSingleTag(key).GetValue() for key in md.GetKeys()}
        return TaggedImage(tags, pix)

    def get_tagged_image(self, cam_index, camera, height, width, binning=None, pixel_type=None, roi_x_start=None,
                         roi_y_start=None):
        """
        Different signature than the Java version because of difference in metadata handling in the swig layers
        """
        pix = self.get_image()
        md = pymmcore.Metadata()
        # most of the same tags from pop_next_tagged_image, which may not be the same as the MMCoreJ version of this function
        tags = {'Camera': camera, 'Height': height, 'Width': width, 'PixelType': pixel_type,
                'CameraChannelIndex': cam_index}
        # Could optionally add these for completeness but there might be a performance hit
        if binning is not None:
            tags['Binning'] = binning
        if roi_x_start is not None:
            tags['ROI-X-start'] = roi_x_start
        if roi_y_start is not None:
            tags['ROI-Y-start'] = roi_y_start

        return TaggedImage(tags, pix)

    instance.get_tagged_image = types.MethodType(get_tagged_image, instance)
    instance.pop_next_tagged_image = types.MethodType(pop_next_tagged_image, instance)

    # attach TaggedImage class
    instance.TaggedImage = TaggedImage
    return instance


_JAVA_HEADLESS_SUBPROCESSES = []
_PYMMCORES = []

def stop_headless(debug=False):
    
    for p in _JAVA_HEADLESS_SUBPROCESSES:
        port = p.port
        if debug:
            logging.main_logger.debug('Stopping headless process with pid {}'.format(p.pid))
        p.terminate()
        server_terminated(port)
        if debug:
            logging.main_logger.debug('Waiting for process with pid {} to terminate'.format(p.pid))
        p.wait()  # wait for process to terminate
        if debug:
            logging.main_logger.debug('Process with pid {} terminated'.format(p.pid))
    _JAVA_HEADLESS_SUBPROCESSES.clear()
    if debug:
        logging.main_logger.debug('Stopping {} pymmcore instances'.format(len(_PYMMCORES)))
    for c in _PYMMCORES:
        if debug:
            logging.main_logger.debug('Stopping pymmcore instance')
        c.unloadAllDevices()
        if debug:
            logging.main_logger.debug('Unloaded all devices')
        Engine.get_instance().shutdown()
        if debug:
            logging.main_logger.debug('Engine shut down')
    _PYMMCORES.clear()
    if debug:
        logging.main_logger.debug('Headless stopped')

# make sure any Java processes are cleaned up when Python exits
atexit.register(stop_headless)

def start_headless(
    mm_app_path: str, config_file: str=None, java_loc: str=None,
        python_backend=False, core_log_path: str='',
        buffer_size_mb: int=1024, max_memory_mb: int=2000,
        port: int=DEFAULT_BRIDGE_PORT, debug=False):
    """
    Start a Java process that contains the neccessary libraries for pycro-manager to run,
    so that it can be run independently of the Micro-Manager GUI/application. This calls
    will create and initialize MMCore with the configuration file provided.

    On windows plaforms, the Java Runtime Environment will be grabbed automatically
    as it is installed along with the Micro-Manager application.

    On non-windows platforms, it may need to be installed/specified manually in order to ensure compatibility.
    Installing Java 11 is the most likely version to work without issue

    Parameters
    ----------
    mm_app_path : str
        Path to top level folder of Micro-Manager installation (made with graphical installer)
    config_file : str
        Path to micro-manager config file, with which core will be initialized. If None then initialization
        is left to the user.
    java_loc: str
        Path to the java version that it should be run with (Java backend only)
    python_backend : bool
        Whether to use the python backend or the Java backend
    core_log_path : str
        Path to where core log files should be created
    buffer_size_mb : int
        Size of circular buffer in MB in MMCore
    max_memory_mb : int
        Maximum amount of memory to be allocated to JVM (Java backend only
    port : int
        Default port to use for ZMQServer (Java backend only)
    debug : bool
        Print debug messages
    """

    if python_backend:
        mmc = _create_pymmcore_instance()
        mmc.set_device_adapter_search_paths([mm_app_path])
        if config_file is not None and config_file != "":
            mmc.load_system_configuration(config_file)
        mmc.set_circular_buffer_memory_footprint(buffer_size_mb)
        _PYMMCORES.append(mmc) # Store so it doesn't get garbage collected
        Engine(mmc)
    else:
        classpath = mm_app_path + '/plugins/Micro-Manager/*'
        if java_loc is None:
            if platform.system() == "Windows":
                # windows comes with its own JRE
                java_loc = mm_app_path + "/jre/bin/javaw.exe"
            else:
                java_loc = "java"
        # This starts Java process and instantiates essential objects (core,
        # acquisition engine, ZMQServer)
        process = subprocess.Popen(
                [
                    java_loc,
                    "-classpath",
                    classpath,
                    "-Dsun.java2d.dpiaware=false",
                    f"-Xmx{max_memory_mb}m",
                    # This is used by MM desktop app but breaks things on MacOS...Don't think its neccessary
                    # "-XX:MaxDirectMemorySize=1000",
                    "org.micromanager.remote.HeadlessLauncher",
                    str(port),
                    config_file if config_file is not None else '',
                    str(buffer_size_mb),
                    core_log_path,
                ], cwd=mm_app_path, stdout=subprocess.PIPE
            )
        process.port = port
        _JAVA_HEADLESS_SUBPROCESSES.append(process)

        started = False
        output = True
        # Some drivers output various status messages which need to be skipped over to look for the STARTED token.
        while output and not started:
            output = process.stdout.readline()
            started = "STARTED" in output.decode('utf-8')
        if not started:
            raise Exception('Error starting headless mode')
        if debug:
            logging.main_logger.debug('Headless mode started')
            def loggerFunction():
                while process in _JAVA_HEADLESS_SUBPROCESSES:
                    line = process.stdout.readline().decode('utf-8')
                    if line.strip() != '':
                        logging.main_logger.debug(line)
            threading.Thread(target=loggerFunction).start()





from pycromanager.headless import _PYMMCORES
from pycromanager.acquisition.java_backend_acquisitions import JavaBackendAcquisition
from pycromanager.acquisition.python_backend_acquisitions import PythonBackendAcquisition
from pycromanager.acquisition.acquisition_superclass import Acquisition as PycromanagerAcquisitionBase
from inspect import signature

# This is a convenience class that automatically selects the appropriate acquisition
# type based on backend is running. It is subclassed from the base acquisition class
# so that it can inherit its docstrings. It cant be the parent class of it, or else
# there will be a circular import
class Acquisition(PycromanagerAcquisitionBase):
    def __new__(cls,
            directory: str = None,
            name: str = 'default_acq_name',
            image_process_fn: callable = None,
            event_generation_hook_fn: callable = None,
            pre_hardware_hook_fn: callable = None,
            post_hardware_hook_fn: callable = None,
            post_camera_hook_fn: callable = None,
            notification_callback_fn: callable = None,
            image_saved_fn: callable = None,
            napari_viewer=None,
            debug: int = False,
                **kwargs):
        # package up all the arguments
        arg_names = [k for k in signature(Acquisition.__init__).parameters.keys() if k != 'self']
        l = locals()
        named_args = {arg_name: (l[arg_name] if arg_name in l else
                                     dict(signature(Acquisition.__init__).parameters.items())[arg_name].default)
                                     for arg_name in arg_names }

        if _PYMMCORES:
            # Python backend detected, so create a python backend acquisition
            specific_arg_names = [k for k in signature(JavaBackendAcquisition.__init__).parameters.keys() if k != 'self']
            for name in specific_arg_names:
                if name in kwargs:
                    named_args[name] = kwargs[name]
            return PythonBackendAcquisition(**named_args)
        else:
            # add any kwargs are specific to java backend
            specific_arg_names = [k for k in signature(JavaBackendAcquisition.__init__).parameters.keys() if k != 'self']
            for name in specific_arg_names:
                if name in kwargs:
                    named_args[name] = kwargs[name]
            return JavaBackendAcquisition(**named_args)
